from copy import deepcopy
import hashlib
from pathlib import Path
from collections.abc import Mapping
import os
import warnings

_KEY = 'resources_'


def _cast_to_path(value, key):
    try:
        return Path(value)
    except TypeError as e:
        raise TypeError(f'Error reading params.resources_ with key {key!r}. '
                        f'Expected value {value!r} to be a str, bytes '
                        f'or os.PathLike, not {type(value).__name__}') from e


def _check_is_file(path, key):
    if not path.is_file():
        raise FileNotFoundError(
            f'Error reading params.resources_ with key {key!r}. '
            f'Expected value {str(path)!r} to be an existing file.')


def _check_file_size(path):
    resource_stat = os.stat(path)
    resource_file_size = resource_stat.st_size / 1E+6
    if resource_file_size > 1:
        warnings.warn(
            f"resource_ {path!r} is {resource_file_size:.1f} MB. "
            "It is not recommended to use large files in "
            "resources_ since it increases task initialization time".format(
                path=path, resource_file_size=resource_file_size))


def _validate(params):
    if not isinstance(params[_KEY], Mapping):
        raise TypeError(
            "Error reading params.resources_. 'resources_' must be a "
            "dictionary with paths to files to track, but got a value "
            f"{params[_KEY]} with type {type(params[_KEY]).__name__}")


def resolve_resources(params, relative_to):
    # params can be None
    if params is None:
        return None

    if _KEY not in params:
        return deepcopy(params)

    _validate(params)

    resources = {}

    for key, value in params[_KEY].items():
        path = _cast_to_path(value, key)
        path_with_suffix = Path(relative_to, path)
        _check_is_file(path_with_suffix, key)
        resources[key] = str(path_with_suffix.resolve())

    params_out = deepcopy(params)
    params_out[_KEY] = resources

    return params_out


def process_resources(params):
    """
    Process resources in a parameters dict, computes the hash of the file for
    resources (i.e., params with the resource__ prefix)

    Parameters
    ----------
    params : dict
        Task parameters
    """

    # params can be None
    if params is None:
        return None

    if _KEY not in params:
        return deepcopy(params)

    _validate(params)

    resources = {}

    for key, value in params[_KEY].items():
        path = _cast_to_path(value, key)
        _check_is_file(path, key)
        _check_file_size(path)
        digest = hashlib.md5(path.read_bytes()).hexdigest()
        resources[key] = digest

    params_out = deepcopy(params)
    params_out[_KEY] = resources

    return params_out



from copy import copy, deepcopy
import importlib
from pathlib import Path
from collections.abc import Mapping
from reprlib import Repr

import yaml
from jinja2 import Template, StrictUndefined

from ploomber.env import validate
from ploomber.env.expand import EnvironmentExpander
from ploomber.env.frozenjson import FrozenJSON
from ploomber import repo
from ploomber.util import default
from ploomber.util.validate import keys as validate_keys
from ploomber.placeholders import util
from ploomber.exceptions import BaseException, ValidationError


def _get_import_from(raw_data, path_to_here):
    meta = raw_data.get('meta', {})

    if not isinstance(meta, Mapping):
        raise ValidationError(
            "Expected 'meta' to contain a "
            f"dictionary, but got: {meta} ({type(meta).__name__})")

    validate_keys(valid={'import_from'}, passed=set(meta), name='meta')

    if 'import_from' in meta:
        import_from = meta['import_from']

        if not isinstance(import_from, str):
            raise ValidationError(
                "Expected 'import_from' to contain a "
                "string, but "
                f"got: {import_from} ({type(import_from).__name__})")

        path = Path(path_to_here, import_from)

        if not path.is_file():
            raise ValidationError(
                "Expected import_from "
                f"value {import_from!r} to be a path to a file, "
                "but such file does not exist")

        return EnvDict(path)._data

    return None


# TODO: custom expanders, this could be done trough another special directive
# such as _expander_class to know which class to use
class EnvDict(Mapping):
    """
    Implements the initialization functionality for Env, except it allows
    to more than one instance to exist, this is used internally to allow
    factory functions introspection without having to create an actual Env

    Parameters
    ----------
    source : dict or str
        If str, it will be interpreted as a path to a YAML file

    path_to_here : str or pathlib.Path
        Value used to expand the {{here}} placeholder. If None, it uses the
        location of the YAML spec. If initialized with a dict and None,
        the {{here}} placeholder is not available.

    defaults : dict, default=None
        Default values to use. If not None, it uses these as defaults and
        overwrites keys using values in source

    Notes
    -----
    By default, it includes the following placeholders (unless the passed
    dictionary already contains those keys): {{user}} (current user)
    {{cwd}} (working directory), {{here}} (env.yaml location, if any), {{root}}
    (project's root folder, if any)
    """
    def __init__(self, source, path_to_here=None, defaults=None):
        # if initialized from another EnvDict, copy the attributes to
        # initialize
        # this happens in the  CLI parser, which instanttiates the env
        # because it needs to create one and then replace cli args, then
        # passes this modified object to DAGSpec
        if isinstance(source, EnvDict):
            for attr in (
                    '_path_to_env',
                    '_preprocessed',
                    '_expander',
                    '_data',
                    '_repr',
                    '_default_keys',
            ):
                original = getattr(source, attr)
                setattr(self, attr, deepcopy(original))
        else:
            (
                # load data
                raw_data,
                # this will be None if source is a dict
                self._path_to_env) = load_from_source(source)

            if defaults:
                raw_data = {**defaults, **raw_data}

            # initialize expander, which converts placeholders to their values
            # we need to pass path_to_env since the {{here}} placeholder
            # resolves to its parent
            if path_to_here is None:
                # if no pat_to_here, use path_to_end
                path_to_here = (None if self._path_to_env is None else Path(
                    self._path_to_env).parent)
            else:
                path_to_here = Path(path_to_here).resolve()

            # add default placeholders but override them if they are defined
            # in the raw data
            default = self._default_dict(path_to_here=path_to_here)
            self._default_keys = set(default) - set(raw_data)
            raw_data = {**default, **raw_data}

            # check if this one is importing
            import_from = _get_import_from(raw_data, path_to_here)

            if import_from:
                raw_data = {**import_from, **raw_data}

            # check raw data is ok
            validate.raw_data_keys(raw_data)

            # expand _module special key, return its expanded value
            self._preprocessed = raw_preprocess(raw_data, self._path_to_env)

            self._expander = EnvironmentExpander(self._preprocessed,
                                                 path_to_here=path_to_here,
                                                 path_to_env=self._path_to_env)
            # now expand all values
            data = self._expander.expand_raw_dictionary(raw_data)
            # this section is for config and should not be visible
            data.pop('meta', None)
            self._data = data

            self._repr = Repr()

    @classmethod
    def find(cls, source):
        """
        Find env file recursively, currently only used by the @with_env
        decorator
        """
        if not Path(source).exists():
            source_found, _ = default.find_file_recursively(source)

            if source_found is None:
                raise FileNotFoundError('Could not find file "{}" in the '
                                        'current working directory nor '
                                        '6 levels up'.format(source))
            else:
                source = source_found

        return cls(source, path_to_here=Path(source).parent)

    @property
    def default_keys(self):
        """Returns keys whose default value is used (i.e., if the user
        overrides them, they won't appear)
        """
        return self._default_keys

    @staticmethod
    def _default_dict(path_to_here):

        placeholders = {
            'user': '{{user}}',
            'cwd': '{{cwd}}',
            'now': '{{now}}',
        }

        if default.try_to_find_root_recursively(path_to_here) is not None:
            placeholders['root'] = '{{root}}'

        if path_to_here is not None:
            placeholders['here'] = '{{here}}'

        if path_to_here is not None and repo.is_repo(path_to_here):
            placeholders['git'] = '{{git}}'
            placeholders['git_hash'] = '{{git_hash}}'

        return placeholders

    @property
    def path_to_env(self):
        return self._path_to_env

    def __getattr__(self, key):
        error = AttributeError("'{}' object has no attribute '{}'".format(
            type(self).__name__, key))
        # do not look up special atttributes this way!
        if key.startswith('__') and key.endswith('__'):
            raise error

        if key in self:
            return self[key]
        else:
            raise AttributeError("{} object has no atttribute '{}'".format(
                repr(self), key))

    def __getitem__(self, key):
        try:
            return self._getitem(key)
        except KeyError as e:
            # custom error will be displayed around quotes, but it's fine.
            # this is due to the KeyError.__str__ implementation
            msg = "{} object has no key '{}'".format(repr(self), key)
            e.args = (msg, )
            raise

    def _getitem(self, key):
        if key in self._preprocessed:
            return FrozenJSON(self._preprocessed[key])
        else:
            return FrozenJSON(self._data[key])

    def __setitem__(self, key, value):
        self._data[key] = value

    def __iter__(self):
        for k in self._data:
            yield k

    def __len__(self):
        return len(self._data)

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        content = self._repr.repr_dict(self._data, level=2)
        return f'{type(self).__name__}({content})'

    def _replace_value(self, value, keys_all):
        """
        Replace a value in the underlying dictionary, by passing a value and
        a list of keys

        e.g. given {'a': {'b': 1}}, we can replace 1 by doing
        _replace_value(2, ['a', 'b'])
        """
        keys_to_final_dict = keys_all[:-1]
        key_to_edit = keys_all[-1]

        dict_to_edit = self._data

        for e in keys_to_final_dict:
            dict_to_edit = dict_to_edit[e]

        if dict_to_edit.get(key_to_edit) is None:
            dotted_path = '.'.join(keys_all)
            raise KeyError('Trying to replace key "{}" in env, '
                           'but it does not exist'.format(dotted_path))

        dict_to_edit[key_to_edit] = (self._expander.expand_raw_value(
            value, keys_all))

    def _inplace_replace_flatten_key(self, value, key_flatten):
        """
        Replace a value in the underlying dictionary, by passing a value and
        a list of keys

        e.g. given {'a': {'b': 1}}, we can replace 1 by doing
        _replace_flatten_keys(2, 'env__a__b'). This function is used
        internally to overrive env values when calling factories (functions
        decorated with @with_env or when doing so via the command line
        interface - ploomber build pipeline.yaml --env--a--b 2)

        Returns a copy
        """
        # convert env__a__b__c -> ['a', 'b', 'c']
        parts = key_flatten.split('__')

        if parts[0] != 'env':
            raise ValueError('keys_flatten must start with env__')

        keys_all = parts[1:]
        self._replace_value(value, keys_all)

    def _replace_flatten_key(self, value, key_flatten):
        obj = copy(self)
        obj._inplace_replace_flatten_key(value, key_flatten)
        return obj

    def _inplace_replace_flatten_keys(self, to_replace):
        """Replace multiple keys at once

        Returns a copy
        """
        for key, value in to_replace.items():
            self._inplace_replace_flatten_key(value, key)

    def _replace_flatten_keys(self, to_replace):
        obj = copy(self)
        obj._inplace_replace_flatten_keys(to_replace)
        return obj

    def _render(self, raw_value):
        placeholders = util.get_tags_in_str(raw_value,
                                            require_runtime_placeholders=False)

        if not placeholders:
            return raw_value, placeholders

        undefined = placeholders - set(self)

        if undefined:
            msg = _error_message_for_undefined_list(undefined)
            raise BaseException(
                f'Error replacing placeholders:\n{msg}\n\nLoaded env: '
                f'{self!r}')

        # render using self._data since self contains FrozenJSON objects
        value = Template(raw_value,
                         undefined=StrictUndefined).render(**self._data)

        return value, placeholders


def _error_message_for_undefined_list(undefined):
    return '\n'.join(_error_message_for_undefined(u) for u in undefined)


def _error_message_for_undefined(u):
    git_ = 'Ensure git is installed and git repository exists'
    reasons = {
        'git':
        git_,
        'git_hash':
        git_,
        'here':
        'Ensure the spec was initialized from a file',
        'root': ('Ensure a pipeline.yaml or setup.py exist in the '
                 'current directory or a parent directory')
    }
    howto = reasons.get(u, 'Ensure the placeholder is defined in the env')

    return '  * {{' + u + '}}: ' + howto


def load_from_source(source):
    """
    Loads from a dictionary or a YAML and applies preprocesssing to the
    dictionary

    Returns
    -------
    dict
        Raw dictionary
    pathlib.Path
        Path to the loaded file, None if source is a dict
    str
        Name, if loaded from a YAML file with the env.{name}.yaml format,
        None if another format or if source is a dict
    """
    if isinstance(source, Mapping):
        # dictiionary, path
        return source, None

    with open(str(source)) as f:
        try:
            raw = yaml.load(f, Loader=yaml.SafeLoader)
        except Exception as e:
            raise type(e)('yaml.load failed to parse your YAML file '
                          'fix syntax errors and try again') from e
        finally:
            # yaml.load returns None for empty files and str if file just
            # contains a string - those aren't valid for our use case, raise
            # an error
            if not isinstance(raw, Mapping):
                raise ValueError("Expected object loaded from '{}' to be "
                                 "a dict but got '{}' instead, "
                                 "verify the content".format(
                                     source,
                                     type(raw).__name__))

    path = Path(source).resolve()

    return raw, path


def raw_preprocess(raw, path_to_raw):
    """
    Preprocess a raw dictionary. If a '_module' key exists, it
    will be expanded: first, try to locate a module with that name and resolve
    to their location (root __init__.py parent), if no module is found,
    interpret as a path to the project's root folder, checks that the folder
    actually exists. '{{here}}' is also allowed, which resolves to the
    path_to_raw, raises Exception if path_to_raw is None

    Returns
    -------
    preprocessed : dict
        Dict with preprocessed keys (empty dictionary it no special
        keys exist in raw)

    Parameters
    ----------
    raw : dict
        Raw data dictionary
    path_to_raw : str
        Path to file where dict was read from, it read from a dict, pass None
    """
    module = raw.get('_module')
    preprocessed = {}

    if module:

        if raw['_module'] == '{{here}}':

            if path_to_raw is not None:
                preprocessed['_module'] = path_to_raw.parent
            else:
                raise ValueError('_module cannot be {{here}} if '
                                 'not loaded from a file')
        else:
            # check if it's a filesystem path
            as_path = Path(module)

            if as_path.exists():
                if as_path.is_file():
                    raise ValueError(
                        'Could not resolve _module "{}", '
                        'expected a module or a directory but got a '
                        'file'.format(module))
                else:
                    path_to_module = as_path

            # must be a dotted path
            else:
                module_spec = importlib.util.find_spec(module)

                # package does not exist
                if module_spec is None:
                    raise ValueError('Could not resolve _module "{}", '
                                     'it is not a valid module '
                                     'nor a directory'.format(module))
                else:
                    path_to_module = Path(module_spec.origin).parent

            preprocessed['_module'] = path_to_module

    return preprocessed



"""
Utilities for dotted paths

Definitions:
    - dotted path:
        a string pointing to an object (e.g. my_module.my_function)
    - dotted path spec:
        a str or dict pointing to an object with optional parameters for
        calling it (a superset of the above)
"""
import warnings
from pathlib import Path
import importlib
from collections.abc import Mapping
from itertools import chain
from functools import reduce

import parso
import pydantic

from ploomber.exceptions import SpecValidationError


class DottedPath:
    """

    Parameters
    ----------
    dotted_path : str or dict
        A dotted path string such as module.function_name or a dictionary
        with a "dotted_path" key and any extra keys to be used as keyword
        arguments when calling the dotted path

    lazy_load : bool, default=False
        If True, defers dotted path loading until __call__ is executed

    allow_return_none : bool, default=True
        If True, it allows calling the dotted path to return None, otherwise
        it raises an exception
    """
    def __init__(self, dotted_path, lazy_load=False, allow_return_none=True):
        self._spec = DottedPathSpecModel.from_spec(dotted_path)
        self._callable = None
        self._allow_return_none = allow_return_none

        if not lazy_load:
            self._load_callable()

    @property
    def callable(self):
        return self._callable

    def _load_callable(self):
        self._callable = load_callable_dotted_path(self._spec.dotted_path)

    def __call__(self, *args, **kwargs):
        if self._callable is None:
            self._load_callable()

        spec_kwargs = self._spec.get_kwargs()
        overlap = set(spec_kwargs) & set(kwargs)

        if overlap:
            overlap_pretty = ", ".join(f"'{w}'" for w in overlap)
            warnings.warn('Got duplicated arguments '
                          f'({overlap_pretty}) when calling dotted path '
                          f'{self._spec.dotted_path!r}. Overriding values...')

        kwargs_final = {**spec_kwargs, **kwargs}

        # FIXME: this overlaps in functionality
        # with call_dotted_path so we must evaluate if we can replace
        # call_dotted_path with DottedPath
        out = self._callable(*args, **kwargs_final)

        if out is None and not self._allow_return_none:
            raise TypeError('Error calling dotted '
                            f'path {self._spec.dotted_path!r}. '
                            'Expected a value but got None')

        return out

    def __repr__(self):
        repr_ = f'{type(self).__name__}({self._spec.dotted_path!r})'

        if self._callable is not None:
            repr_ += f' (loaded: {self._callable})'

        return repr_


def _validate_dotted_path(dotted_path, raise_=True):
    parts = dotted_path.split('.')

    if len(parts) < 2 or not all(parts):
        if raise_:
            raise ValueError(f'Invalid dotted path {dotted_path!r}. '
                             'Value must be a dot separated '
                             'string, with at least two parts: '
                             '[module_name].[function_name]')
        else:
            return False

    return '.'.join(parts[:-1]), parts[-1]


def load_dotted_path(dotted_path, raise_=True, reload=False):
    """Load an object/function/module by passing a dotted path

    Parameters
    ----------
    dotted_path : str
        Dotted path to a module, e.g. ploomber.tasks.NotebookRunner
    raise_ : bool, default=True
        If True, an exception is raised if the module can't be imported,
        otherwise return None if that happens
    reload : bool, default=False
        Reloads the module after importing it
    """
    obj, module = None, None

    parsed = _validate_dotted_path(dotted_path, raise_=raise_)

    if parsed:
        mod, name = parsed
        main_mod = str(mod.split('.')[0])
        try:
            module = importlib.import_module(mod)
        except ModuleNotFoundError as e:
            if raise_:
                spec = importlib.util.find_spec(main_mod)

                msg = ('An error occured when trying to import dotted '
                       f'path {dotted_path!r}: {e}')

                if spec is not None:
                    msg = (msg +
                           f' (loaded {main_mod!r} from {spec.origin!r})')

                e.msg = msg

                raise

        if module:
            if reload:
                module = importlib.reload(module)

            try:
                obj = getattr(module, name)
            except AttributeError as e:
                if raise_:
                    e.args = ((f'Could not get {name!r} from module {mod!r} '
                               f'(loaded {mod!r} from {module.__file__!r}). '
                               'Ensure it is defined in such module'), )
                    raise
        return obj
    else:
        if raise_:
            raise ValueError(
                'Invalid dotted path value "{}", must be a dot separated '
                'string, with at least '
                '[module_name].[function_name]'.format(dotted_path))


def load_callable_dotted_path(dotted_path, raise_=True, reload=False):
    """
    Like load_dotted_path but verifies the loaded object is a callable
    """
    loaded_object = load_dotted_path(dotted_path=dotted_path,
                                     raise_=raise_,
                                     reload=reload)

    if not callable(loaded_object):
        raise TypeError(f'Error loading dotted path {dotted_path!r}. '
                        'Expected a callable object (i.e., some kind '
                        f'of function). Got {loaded_object!r} '
                        f'(an object of type: {type(loaded_object).__name__})')

    return loaded_object


def call_dotted_path(dotted_path, raise_=True, reload=False, kwargs=None):
    """
    Load dotted path (using load_callable_dotted_path), and call it with
    kwargs arguments, raises an exception if returns None

    Parameters
    ----------
    dotted_path : str
        Dotted path to call
    kwargs : dict, default=None
        Keyword arguments to call the dotted path
    """
    callable_ = load_callable_dotted_path(dotted_path=dotted_path,
                                          raise_=raise_,
                                          reload=reload)

    kwargs = kwargs or dict()

    try:
        out = callable_(**kwargs)
    except Exception as e:
        origin = locate_dotted_path(dotted_path).origin
        msg = str(e) + f' (Loaded from: {origin})'
        e.args = (msg, )
        raise

    if out is None:
        raise TypeError(f'Error calling dotted path {dotted_path!r}. '
                        'Expected a value but got None')

    return out


# TODO: this must be something like "locate_dotted_path_spec" since it's
# returning that
def locate_dotted_path(dotted_path):
    """
    Locates a dotted path, returns the spec for the module where the attribute
    is defined
    """
    tokens = dotted_path.split('.')
    module = '.'.join(tokens[:-1])
    # NOTE: if importing a sub-module (e.g., something.another), this will
    # import some modules (rather than just locating them) - I think we
    # should remove them to pervent import clashes
    spec = importlib.util.find_spec(module)

    if spec is None:
        raise ModuleNotFoundError(f'Module {module!r} does not exist')

    return spec


def locate_dotted_path_root(dotted_path):
    """
    Returns the module spec for a given dotted path.
    e.g. module.sub.another, checks that module exists
    """
    tokens = dotted_path.split('.')
    spec = importlib.util.find_spec(tokens[0])

    if spec is None:
        raise ModuleNotFoundError(f'Module {tokens[0]!r} does not exist')

    return spec


def _process_children(ch):
    if hasattr(ch, 'name'):
        return [(ch.name.value, ch.type, ch.get_code())]
    else:
        nested = ((c.get_defined_names(), c.type, c.get_code().strip())
                  for c in ch.children if hasattr(c, 'get_defined_names'))
        return ((name.value, type_, code) for names, type_, code in nested
                for name in names)


def _check_last_definition_is_function(module, name, dotted_path):
    children = [
        ch for ch in module.children[::-1]
        if hasattr(ch, 'name') or hasattr(ch, 'children')
    ]

    gen = chain(*(_process_children(ch) for ch in children))

    last_type = None
    code = None

    for name_, type_, code in gen:
        if name_ == name:
            last_type = type_
            last_code = code
            break

    if last_type and last_type != 'funcdef':
        raise TypeError(f'Failed to load dotted path {dotted_path!r}. '
                        f'Expected last defined {name!r} to be a function. '
                        f'Got:\n{last_code!r}')


def _check_defines_function_with_name(path, name, dotted_path):
    module = parso.parse(Path(path).read_text())

    # there could be multiple with the same name
    fns = [fn for fn in module.iter_funcdefs() if fn.name.value == name]

    if not fns:
        # check if there are import statements defining the attribute
        imports = [
            imp for imp in module.iter_imports()
            if name in [n.value for n in imp.get_defined_names()]
        ]

        if imports:
            # FIXME: show all imports in the error message instead of
            # only the first one
            import_ = imports[0].get_code()

            raise NotImplementedError(
                'Failed to locate dotted path '
                f'{dotted_path!r}, definitions from import statements are not '
                f'supported. Move the defitinion of function {name!r} '
                f'to {str(path)!r} and delete the import statement {import_!r}'
            )

        raise AttributeError(
            f'Failed to locate dotted path {dotted_path!r}. '
            f'Expected {str(path)!r} to define a function named {name!r}')

    # return the last definition to be consistent with inspect.getsourcefile
    fn_found = fns[-1]

    # if function is decorated, include decorators in the returned source code
    if fn_found.parent.type == 'decorated':
        fn_found = fn_found.parent

    _check_last_definition_is_function(module, name, dotted_path)

    return f'{path}:{fn_found.start_pos[0]}', fn_found.get_code().lstrip()


def lazily_locate_dotted_path(dotted_path):
    """
    Locates a dotted path, but unlike importlib.util.find_spec, it does not
    import submodules
    """
    _validate_dotted_path(dotted_path)
    parts = dotted_path.split('.')

    module_name = '.'.join(parts[:-1])
    first, middle, mod, symbol = parts[0], parts[1:-2], parts[-2], parts[-1]

    spec = importlib.util.find_spec(first)

    if spec is None:
        raise ModuleNotFoundError('Error processing dotted '
                                  f'path {dotted_path!r}, '
                                  f'no module named {first!r}')

    # python 3.6 returns 'namespace', python 3.7 an up returns None
    if spec.origin is None or spec.origin == 'namespace':
        raise ModuleNotFoundError('Error processing dotted '
                                  f'path {dotted_path!r}: '
                                  f'{first!r} appears to be a namespace '
                                  'package, which are not supported')

    origin = Path(spec.origin)
    location = origin.parent

    # a.b.c.d.e.f
    # a/__init__.py or a.py must exist
    # from b until d, there must be {name}/__init__.py
    # there must be e/__init__.py or e.py
    # f must be a symbol defined at e.py or e/__init__.py

    if len(parts) == 2:
        return _check_defines_function_with_name(origin, symbol, dotted_path)

    location = reduce(lambda x, y: x / y, [location] + middle)

    init = (location / mod / '__init__.py')
    file_ = (location / f'{mod}.py')

    if init.exists():
        return _check_defines_function_with_name(init, symbol, dotted_path)
    elif file_.exists():
        return _check_defines_function_with_name(file_, symbol, dotted_path)
    else:
        raise ModuleNotFoundError(f'No module named {module_name!r}. '
                                  f'Expected to find one of {str(init)!r} or '
                                  f'{str(file_)!r}, but none of those exist')


def dotted_path_exists(dotted_path):
    try:
        return lazily_locate_dotted_path(dotted_path)
    except (ModuleNotFoundError, AttributeError):
        return False


class BaseModel(pydantic.BaseModel):
    """Base model for specs
    """
    def __init__(self, **kwargs):
        # customize ValidationError message
        try:
            super().__init__(**kwargs)
        except pydantic.ValidationError as e:
            ex = e
        else:
            ex = None

        if ex:
            raise SpecValidationError(errors=ex.errors(),
                                      model=type(self),
                                      kwargs=kwargs)


class DottedPathSpecModel(BaseModel):
    dotted_path: str

    class Config:
        extra = 'allow'

    @classmethod
    def from_spec(cls, dotted_path_spec):
        if isinstance(dotted_path_spec, str):
            return cls(dotted_path=dotted_path_spec)
        elif isinstance(dotted_path_spec, Mapping):
            return cls(**dotted_path_spec)
        else:
            raise TypeError(
                'Expected dotted path spec to be a str or Mapping, '
                f'got {dotted_path_spec!r} '
                f'(type: {type(dotted_path_spec).__name__})')

    def get_kwargs(self):
        return self.dict(exclude={'dotted_path'})


def create_intermediate_modules(module_parts):
    """
    Creates the folder structure needed for a module specified
    by the parts of its name

    Parameters
    ----------
    module_parts : list
        A list of strings with the module elements.
        Example: ['module', 'sub_module']

    Raises
    ------
    ValueError
        If the module already exists
    """
    dotted_path_to_module = '.'.join(module_parts)
    *inner, last = module_parts

    # check if it already exists
    if len(module_parts) >= 2:
        fn_check = dotted_path_exists
    else:
        fn_check = importlib.util.find_spec

    if fn_check(dotted_path_to_module):
        raise ValueError(f'Module {dotted_path_to_module!r} already exists')

    # if the root module already exists, we should create the missing files
    # in the existing location
    spec = importlib.util.find_spec(module_parts[0])

    # .origin will be None for namespace packages
    # on Python 3.6, spec.origin is 'namespace' instead of None
    if spec and spec.origin is not None and spec.origin != 'namespace':
        inner[0] = Path(spec.origin).parent

    parent = Path(*inner)
    parent.mkdir(parents=True, exist_ok=True)

    for idx in range(len(inner)):
        init_file = Path(*inner[:idx + 1], "__init__.py")

        if not init_file.exists():
            init_file.touch()

    Path(parent, f"{last}.py").touch()



"""Efficiently upload/download data
"""
from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path

import requests

try:
    import boto3
except ModuleNotFoundError:
    boto3 = None


def yield_index(file_size, max_size):
    parts = int(file_size / max_size)

    for i in range(parts):
        idx = max_size * i

        yield idx, idx + max_size

    if file_size % max_size:
        idx = parts * max_size
        yield idx, file_size


def read_from_index(path, i, j):
    with Path(path).open('rb') as f:
        f.seek(i)
        return f.read(j - i)


def yield_parts(path, max_size):
    file_size = Path(path).stat().st_size

    for i, j in yield_index(file_size, max_size):
        yield read_from_index(path, i, j)


def n_parts(path, max_size=None):
    max_size = max_size or (10 * 1024 * 1024)
    file_size = Path(path).stat().st_size
    return int(file_size / max_size) + int(file_size % max_size > 0)


def generate_links(bucket_name, key, upload_id, n_parts):
    s3 = boto3.client('s3')
    return [
        s3.generate_presigned_url(ClientMethod='upload_part',
                                  Params={
                                      'Bucket': bucket_name,
                                      'Key': key,
                                      'UploadId': upload_id,
                                      'PartNumber': part_no
                                  }) for part_no in range(1, n_parts + 1)
    ]


class UploadJobGenerator:
    """

    Examples
    --------
    >>> from ploomber.cloud import io
    >>> io.upload('v2.mov', 5 * 1024 * 1024,
    ...     'ploomber-bucket', 'raw/v2.mov') # doctest: +SKIP

    Notes
    -----
    https://github.com/boto/boto3/issues/2305
    """
    def __init__(self,
                 path,
                 key,
                 upload_id,
                 links,
                 bucket=None,
                 max_size=None):
        # only reuired if calling complete
        self._bucket = bucket
        self._upload_id = upload_id
        self._max_size = max_size or (10 * 1024 * 1024)
        self._path = path
        self._key = key
        self._links = links
        self._n_parts = n_parts(path, max_size)

    @property
    def n_parts(self):
        return self._n_parts

    @classmethod
    def from_scratch(cls, path, max_size, bucket, key):
        s3 = boto3.client('s3')
        res = s3.create_multipart_upload(Bucket=bucket, Key=key)
        upload_id = res['UploadId']
        links = generate_links(bucket, key, upload_id, n_parts(path, max_size))
        return cls(path, max_size, bucket, key, upload_id, links)

    def __iter__(self):
        file_size = Path(self._path).stat().st_size

        for num, (link, (i, j)) in enumerate(zip(
                self._links, yield_index(file_size, self._max_size)),
                                             start=1):
            yield UploadJob(self._path, link, i, j, num)

    def upload(self, complete=False):

        with ThreadPoolExecutor(max_workers=4) as executor:
            future2job = {
                executor.submit(upload_job): upload_job
                for upload_job in self
            }

            for future in as_completed(future2job):
                exception = future.exception()

                if exception:
                    job = future2job[future]
                    raise RuntimeError(
                        'An error occurred when downloading product from '
                        f'job: {job!r}') from exception

        parts = [job._res for job in future2job.values()]

        if complete:
            self.complete(parts)

        return parts

    def complete(self, parts):
        s3 = boto3.client('s3')
        return s3.complete_multipart_upload(Bucket=self._bucket,
                                            Key=self._key,
                                            MultipartUpload={'Parts': parts},
                                            UploadId=self._upload_id)


class UploadJob:
    def __init__(self, path, link, i, j, num):
        self._path = path
        self._link = link
        self._i = i
        self._j = j
        self._num = num
        self._res = None

    def __call__(self):
        res = requests.put(self._link,
                           data=read_from_index(self._path, self._i, self._j))
        etag = res.headers['ETag']
        self._res = {'ETag': etag, 'PartNumber': self._num}



import re
import warnings
from io import StringIO

import parso
from pyflakes import api as pyflakes_api
from pyflakes.reporter import Reporter
from pyflakes.messages import (UndefinedName, UndefinedLocal,
                               DuplicateArgument, ReturnOutsideFunction,
                               YieldOutsideFunction, ContinueOutsideLoop,
                               BreakOutsideLoop)

from ploomber.exceptions import RenderError
from ploomber.io import pretty_print
from ploomber.sources.nb_utils import find_cell_with_tag

# messages: https://github.com/PyCQA/pyflakes/blob/master/pyflakes/messages.py
_ERRORS = (
    UndefinedName,
    UndefinedLocal,
    DuplicateArgument,
    ReturnOutsideFunction,
    YieldOutsideFunction,
    ContinueOutsideLoop,
    BreakOutsideLoop,
)

_IS_IPYTHON_CELL_MAGIC = r'^\s*%{2}[a-zA-Z]+'
_IS_IPYTHON_LINE_MAGIC = r'^\s*%{1}[a-zA-Z]+'
_IS_INLINE_SHELL = r'^\s*!{1}.+'

HAS_INLINE_PYTHON = {'%%capture', '%%timeit', '%%time', '%time', '%timeit'}


def _process_messages(mesages):
    return '\n'.join(str(msg) for msg in mesages)


def process_errors_and_warnings(messages):
    errors, warnings = [], []

    for message in messages:
        # By default, script templates include an
        # 'upstream = None' line in the parameters cell but users may delete
        # it. If that happens and they reference 'upstream' in the body of the
        # notebook but the task does not have upstream dependencies, the
        # injected cell won't add the upstream variable, causing an undefined
        # name error. To provide more context, we modify the original error and
        # add a hint. Note that this only happens when using the Python API
        # directly, since the Spec API requires an upstream = None placeholder
        if (isinstance(message, UndefinedName)
                and message.message_args == ('upstream', )):
            message.message = (
                message.message +
                '. Did you forget to declare upstream dependencies?')

        if isinstance(message, _ERRORS):
            errors.append(message)
        else:
            warnings.append(message)

    return _process_messages(errors), _process_messages(warnings)


# https://github.com/PyCQA/pyflakes/blob/master/pyflakes/reporter.py
class MyReporter(Reporter):
    def __init__(self):
        self._stdout = StringIO()
        self._stderr = StringIO()
        self._stdout_raw = []
        self._unexpected = False
        self._syntax = False

    def flake(self, message):
        self._stdout_raw.append(message)
        self._stdout.write(str(message))
        self._stdout.write('\n')

    def unexpectedError(self, *args, **kwargs):
        """pyflakes calls this when ast.parse raises an unexpected error
        """
        self._unexpected = True
        return super().unexpectedError(*args, **kwargs)

    def syntaxError(self, *args, **kwargs):
        """pyflakes calls this when ast.parse raises a SyntaxError
        """
        self._syntax = True
        return super().syntaxError(*args, **kwargs)

    def _seek_zero(self):
        self._stdout.seek(0)
        self._stderr.seek(0)

    def _make_error_message(self, error):
        return ('An error happened when checking the source code. '
                f'\n{error}\n\n'
                '(if you want to disable this check, set '
                'static_analysis to "disable" in the task declaration)')

    def _check(self, raise_):
        self._seek_zero()

        # syntax errors are stored in _stderr
        # https://github.com/PyCQA/pyflakes/blob/master/pyflakes/api.py

        error_message = '\n'.join(self._stderr.readlines())

        if self._syntax:
            msg = self._make_error_message(error_message)

            if raise_:
                raise SyntaxError(msg)
            else:
                warnings.warn(msg)

        elif self._unexpected:
            warnings.warn('An unexpected error happened '
                          f'when analyzing code: {error_message.strip()!r}')
        else:
            errors, warnings_ = process_errors_and_warnings(self._stdout_raw)

            if warnings_:
                warnings.warn(warnings_)

            if errors:
                msg = self._make_error_message(errors)

                if raise_:
                    raise RenderError(msg)
                else:
                    warnings.warn(msg)


def check_notebook(nb, params, filename, raise_=True, check_signature=True):
    """
    Perform static analysis on a Jupyter notebook code cell sources

    Parameters
    ----------
    nb : NotebookNode
        Notebook object. Must have a cell with the tag "parameters"

    params : dict
        Parameter that will be added to the notebook source

    filename : str
        Filename to identify pyflakes warnings and errors

    raise_ : bool, default=True
        If True, raises an Exception if it encounters errors, otherwise a
        warning

    Raises
    ------
    SyntaxError
        If the notebook's code contains syntax errors

    TypeError
        If params and nb do not match (unexpected or missing parameters)

    RenderError
        When certain pyflakes errors are detected (e.g., undefined name)
    """
    params_cell, _ = find_cell_with_tag(nb, 'parameters')
    check_source(nb, raise_=raise_)

    if check_signature:
        check_params(params, params_cell['source'], filename, warn=not raise_)


def check_source(nb, raise_=True):
    """
    Run pyflakes on a notebook, wil catch errors such as missing passed
    parameters that do not have default values
    """
    # concatenate all cell's source code in a single string
    source_code = '\n'.join([
        _comment_if_ipython_magic(c['source']) for c in nb.cells
        if c.cell_type == 'code'
    ])

    # this objects are needed to capture pyflakes output
    reporter = MyReporter()

    # run pyflakes.api.check on the source code
    pyflakes_api.check(source_code, filename='', reporter=reporter)

    reporter._check(raise_)


def _comment(line):
    """Comments a line
    """
    return f'# {line}'


def _comment_if_ipython_magic(source):
    """Comments lines into comments if they're IPython magics (cell level)
    """
    # TODO: support for nested cell magics. e.g.,
    # %%timeit
    # %%timeit
    # something()
    lines_out = []
    comment_rest = False

    # TODO: inline magics should add a comment at the end of the line, because
    # the python code may change the dependency structure. e.g.,
    # %timeit z = x + y -> z = x + y # [magic] %timeit
    # note that this only applies to inline magics that take Python code as arg

    # NOTE: magics can take inputs but their outputs ARE NOT saved. e.g.,
    # %timeit x = y + 1
    # running such magic requires having y but after running it, x IS NOT
    # declared. But this is magic dependent %time x = y + 1 will add x to the
    # scope

    for line in source.splitlines():
        cell_magic = _is_ipython_cell_magic(line)

        if comment_rest:
            lines_out.append(_comment(line))
        else:
            line_magic = _is_ipython_line_magic(line)

            # if line magic, comment line
            if line_magic:
                lines_out.append(_comment(line))

            # if inline shell, comment line
            elif _is_inline_shell(line):
                lines_out.append(_comment(line))

            # if cell magic, comment line
            elif cell_magic in HAS_INLINE_PYTHON:
                lines_out.append(_comment(line))

            # if cell magic whose content *is not* Python, comment line and
            # all the remaining lines in the cell
            elif cell_magic:
                lines_out.append(_comment(line))
                comment_rest = True

            # otherwise, don't do anything
            else:
                lines_out.append(line)

    return '\n'.join(lines_out)


def _is_ipython_line_magic(line):
    """
    Determines if the source line is an IPython magic. e.g.,

    %%bash
    for i in 1 2 3; do
      echo $i
    done
    """
    return re.match(_IS_IPYTHON_LINE_MAGIC, line) is not None


def _is_inline_shell(line):
    return re.match(_IS_INLINE_SHELL, line) is not None


def _is_ipython_cell_magic(source):
    """Determines if the source is an IPython cell magic. e.g.,

    %cd some-directory
    """
    m = re.match(_IS_IPYTHON_CELL_MAGIC, source.lstrip())

    if not m:
        return False

    return m.group()


def check_params(passed, params_source, filename, warn=False):
    """
    Check that parameters passed to the notebook match the ones defined
    in the parameters variable

    Parameters
    ----------
    passed : iterable
        Paramters passed to the notebook (params argument)

    params_source : str
        Parameters cell source code

    filename : str
        The task's filename. Only used for displaying in the error message

    warn : bool
        If False, it raises a TypeError if params do not match. If True,
        it displays a warning instead.

    Raises
    ------
    TypeError
        If passed parameters do not match variables declared in params_source
        and warn is False
    """
    # dot not complain if product or upstream are missing since
    # they are not user-defined params
    IGNORE = {'product', 'upstream'}

    declared = _get_defined_variables(params_source) - IGNORE
    passed = set(passed) - IGNORE
    missing = declared - passed
    unexpected = passed - declared

    if missing or unexpected:
        errors = []

        if missing:
            errors.append(f'Missing params: {pretty_print.iterable(missing)} '
                          '(to fix this, pass '
                          f'{pretty_print.them_or_name(missing)} in '
                          'the \'params\' argument)')
        if unexpected:
            first = list(unexpected)[0]
            errors.append(
                'Unexpected '
                f'params: {pretty_print.iterable(unexpected)} (to fix this, '
                f'add {pretty_print.them_or_name(unexpected)} to the '
                '\'parameters\' cell and assign the value as '
                f'None. e.g., {first} = None)')

        msg = (f"Parameters "
               "declared in the 'parameters' cell do not match task "
               f"params. {pretty_print.trailing_dot(errors)} To disable this "
               "check, set 'static_analysis' to 'disable' in the "
               "task declaration.")

        if warn:
            warnings.warn(msg)
        else:
            raise TypeError(
                f"Error rendering notebook {str(filename)!r}. {msg}")


def _get_defined_variables(params_source):
    """
    Return the variables defined in a given source. If a name is defined more
    than once, it uses the last definition. Ignores anything other than
    variable assignments (e.g., function definitions, exceptions)
    """
    used_names = parso.parse(params_source).get_used_names()
    return set(key for key, value in used_names.items()
               if value[-1].is_definition()
               and value[-1].get_definition().type == 'expr_stmt')



from pathlib import Path
from typing import Mapping

import yaml
from pydantic import BaseModel


class TaskResource(BaseModel):
    vcpus: int = None
    memory: int = None
    gpu: int = None

    class Config:
        extra = 'forbid'


class CloudConfig(BaseModel):
    task_resources: Mapping[str, TaskResource] = None

    class Config:
        extra = 'forbid'


def validate():
    path = Path('cloud.yaml')

    # cloud.yaml is optional
    if not path.is_file():
        return

    cfg = yaml.safe_load(path.read_text())
    CloudConfig(**cfg)



"""
This module contains code for lexical analysis and code parsing. It is used
to automatically extract "upstream" and "product" variables from notebooks.

Currently, Python parsing is delegated to the parso library. R parsing is done
via our own lexer and parser. Given that we only need to parse certain
statements, it does aim to implement a full R parser.

Our parsers only have to recognize a simple program, in Python, it is the
following:

upstream = ['a', 'b', 'c']
product = {'d': 'path/to/d.csv', 'e': 'path/to/e.csv'}

In R, this would be (note: <- and = are valid assignment operator):

upstream <- list('a', 'b', 'c')
product <- list(d='path/to/d.csv', e='path/to/d.csv')

The parser should be able to convert the R code above into the Python
representation.
"""



"""
A product representing a File in the local filesystem with an optional client,
which allows the file to be retrieved or backed up in remote storage
"""
import json
import shutil
import os
from pathlib import Path
from reprlib import Repr

from ploomber.products.product import Product
from ploomber.placeholders.placeholder import Placeholder
from ploomber.constants import TaskStatus
from ploomber.products._remotefile import (_RemoteFile,
                                           _fetch_metadata_from_file_product)
from ploomber.products.mixins import ProductWithClientMixin
from ploomber.exceptions import MissingClientError


class File(ProductWithClientMixin, os.PathLike, Product):
    """A file (or directory) in the local filesystem

    Parameters
    ----------
    identifier: str or pathlib.Path
        The path to the file (or directory), can contain placeholders
        (e.g. {{placeholder}})
    """
    def __init__(self, identifier, client=None):
        super().__init__(identifier)
        self._client = client
        self._repr = Repr()
        self._repr.maxstring = 40
        self._remote_ = _RemoteFile(self)

    def _init_identifier(self, identifier):
        if not isinstance(identifier, (str, Path)):
            raise TypeError('File must be initialized with a str or a '
                            'pathlib.Path')

        return Placeholder(str(identifier))

    @property
    def _path_to_file(self):
        return Path(str(self._identifier))

    @property
    def _path_to_metadata(self):
        name = f'.{self._path_to_file.name}.metadata'
        return self._path_to_file.with_name(name)

    @property
    def _remote(self):
        """
        RemoteFile for this File. Returns None if a
        File.client doesn't exist, remote file doesn't exist or remote
        metadata doesn't exist
        """
        return self._remote_

    @property
    def _remote_path_to_metadata(self):
        return self._remote._path_to_metadata

    def fetch_metadata(self):
        # migrate metadata file to keep compatibility with ploomber<0.10
        old_name = Path(str(self._path_to_file) + '.source')
        if old_name.is_file():
            shutil.move(old_name, self._path_to_metadata)

        return _fetch_metadata_from_file_product(self, check_file_exists=True)

    def save_metadata(self, metadata):
        self._path_to_metadata.write_text(json.dumps(metadata))

    def _delete_metadata(self):
        if self._path_to_metadata.exists():
            os.remove(str(self._path_to_metadata))

    def exists(self):
        return self._path_to_file.exists()

    def delete(self, force=False):
        # force is not used for this product but it is left for API
        # compatibility
        if self.exists():
            self.logger.debug('Deleting %s', self._path_to_file)
            if self._path_to_file.is_dir():
                shutil.rmtree(str(self._path_to_file))
            else:
                os.remove(str(self._path_to_file))
        else:
            self.logger.debug('%s does not exist ignoring...',
                              self._path_to_file)

    def __repr__(self):
        # do not shorten, we need to process the actual path
        path = Path(self._identifier.best_repr(shorten=False))

        # if absolute, try to show a shorter version, if possible
        if path.is_absolute():
            try:
                path = path.relative_to(Path('.').resolve())
            except ValueError:
                # happens if the path is not a file/folder within the current
                # working directory
                pass

        content = self._repr.repr(str(path))
        return f'{type(self).__name__}({content})'

    def _check_is_outdated(self, outdated_by_code):
        """
        Unlike other Product implementation that only have to check the
        current metadata, File has to check if there is a metadata remote copy
        and download it to decide outdated status, which yield to task
        execution or product downloading
        """
        should_download = False

        if self._remote.exists():
            if self._remote._is_equal_to_local_copy():
                return self._remote._is_outdated(with_respect_to_local=True)
            else:
                # download when doing so will bring the product
                # up-to-date (this takes into account upstream
                # timestamps)
                should_download = not self._remote._is_outdated(
                    with_respect_to_local=True,
                    outdated_by_code=outdated_by_code)

        if should_download:
            return TaskStatus.WaitingDownload

        # no need to download, check status using local metadata
        return super()._check_is_outdated(outdated_by_code=outdated_by_code)

    def _is_remote_outdated(self, outdated_by_code):
        """
        Check status using remote metadata, if no remote is available
        (or remote metadata is corrupted) returns True
        """
        if self._remote.exists():
            return self._remote._is_outdated(with_respect_to_local=False,
                                             outdated_by_code=outdated_by_code)
        else:
            # if no remote, return True. This is the least destructive option
            # since we don't know what will be available and what not when this
            # executes
            return True

    @property
    def client(self):
        try:
            client = super().client
        except MissingClientError:
            return None
        else:
            return client

    def download(self):
        self.logger.info('Downloading %s...', self._path_to_file)

        if self.client:
            self.client.download(str(self._path_to_file))
            self.client.download(str(self._path_to_metadata))

    def upload(self):
        if self.client:
            if not self._path_to_metadata.exists():
                raise RuntimeError(
                    f'Error uploading product {self!r}. '
                    f'Metadata {str(self._path_to_metadata)!r} does '
                    'not exist')

            if not self._path_to_file.exists():
                raise RuntimeError(f'Error uploading product {self!r}. '
                                   f'Product {str(self._path_to_file)!r} does '
                                   'not exist')

            self.logger.info('Uploading %s...', self._path_to_file)
            self.client.upload(self._path_to_metadata)
            self.client.upload(self._path_to_file)

    def __fspath__(self):
        """
        Abstract method defined in the os.PathLike interface, enables this
        to work: ``import pandas as pd; pd.read_csv(File('file.csv'))``
        """
        return str(self)

    def __eq__(self, other):
        return Path(str(self)).resolve() == Path(str(other)).resolve()

    def __hash__(self):
        return hash(Path(str(self)).resolve())



from inspect import isclass
from collections.abc import MutableMapping

from ploomber.tasks.abc import Task
from ploomber.products.product import Product
from ploomber.validators.string import get_suggestion, str_to_class
from ploomber.util.dotted_path import DottedPath


class DAGClients(MutableMapping):
    """
    A dict-like object with validations
    1. __setitem__, __getitem__ work with strings (e.g., clients['SQLScript'])
    2. __setitem__ validates the key is a Task or Product subclass
    """
    def __init__(self, mapping=None):
        self._mapping = mapping or dict()

    def __getitem__(self, key):
        if isinstance(key, str):
            key_obj = str_to_class(key)
        else:
            key_obj = key

        if key_obj is None:
            error = repr(key)

            suggestion = get_suggestion(key)

            if suggestion and str_to_class(suggestion) in self:
                error += f'. Did you mean {suggestion!r}?'

            raise KeyError(error)

        value = self._mapping[key_obj]

        # this happens when loading DAGSpec with lazy_load turned on,
        # clients are not initialized but passed as DottedPath
        if isinstance(value, DottedPath):
            client = value()
            self[key] = client
            return client
        else:
            return value

    def __setitem__(self, key, value):
        if isinstance(key, str):
            key_obj = str_to_class(key)

            if key_obj is None:
                maybe = get_suggestion(key)

                msg = (f'Could not set DAG-level client {value!r}. '
                       f'{key!r} is not a valid Task or '
                       'Product class name')

                if maybe:
                    msg += f'. Did you mean {maybe!r}?'

                raise ValueError(msg)
        else:
            key_obj = key

        if not isclass(key_obj) or not issubclass(key_obj, (Task, Product)):
            raise ValueError('DAG client keys must be Tasks '
                             f'or Products, value {key_obj!r} is not')

        self._mapping[key_obj] = value

    def __delitem__(self, key):
        del self._mapping[key]

    def __iter__(self):
        for item in self._mapping:
            yield item

    def __len__(self):
        return len(self._mapping)

    def __repr__(self):
        return f'{type(self).__name__}({self._mapping!r})'



from pathlib import PurePosixPath, Path

try:
    from google.cloud import storage
except ImportError:
    storage = None

from ploomber.util.default import find_root_recursively
from ploomber.util.util import requires
from ploomber.clients.storage.abc import AbstractStorageClient
from ploomber.exceptions import RemoteFileNotFound, DAGSpecInvalidError


class GCloudStorageClient(AbstractStorageClient):
    """Client for uploading File products to Google Cloud Storage

    Parameters
    ----------
    bucket_name : str
        Bucket to use

    parent : str
        Parent folder in the bucket to store files. For example, if
        ``parent='path/to'``, and a product in your pipeline is
        ``out/data.csv``, your file will appea in the bucket at
        ``path/to/out/data.csv``.

    json_credentials_path : str, default=None
        Use the given JSON file to authenticate the client
        (uses  ``Client.from_service_account_json(**kwargs)``), if ``None``,
        initializes the client using ``Client(**kwargs)``

    path_to_project_root : str, default=None
        Path to project root. If None, looks it
        up automatically and assigns it to the parent folder of your
        ``pipeline.yaml`` spec or ``setup.py`` (if your project is a package).
        This determines the path in remote storage. For example, if
        ``path_to_project_root`` is ``/my-project``, you're storing a
        product at ``/my-project/out/data.csv``, and ``parent='some-dir'``,
        the file will be stored in the bucket at ``some-dir/out/data.csv``
        (we first compute the path of your product relative to the project
        root, then prefix it with ``parent``).

    credentials_relative_to_project_root : bool, default=True
        If ``True``, relative paths in ``json_credentials_path`` are so to the
        ``path_to_project_root``, instead of the current working directory

    **kwargs
        Keyword arguments for the client constructor
    """

    @requires(['google.cloud.storage'],
              name='GCloudStorageClient',
              pip_names=['google-cloud-storage'])
    def __init__(self,
                 bucket_name,
                 parent,
                 json_credentials_path=None,
                 path_to_project_root=None,
                 credentials_relative_to_project_root=True,
                 **kwargs):
        if path_to_project_root:
            project_root = path_to_project_root
        else:
            try:
                project_root = find_root_recursively()
            except Exception as e:
                raise DAGSpecInvalidError(
                    f'Cannot initialize {type(self).__name__} because there '
                    'is not project root. Set one or explicitly pass '
                    'a value in the path_to_project_root argument') from e

        self._path_to_project_root = Path(project_root).resolve()

        if (credentials_relative_to_project_root and json_credentials_path
                and not Path(json_credentials_path).is_absolute()):
            json_credentials_path = Path(self._path_to_project_root,
                                         json_credentials_path)

        self._from_json = json_credentials_path is not None

        if not self._from_json:
            self._client_kwargs = kwargs
        else:
            self._client_kwargs = {
                'json_credentials_path': json_credentials_path,
                **kwargs
            }

        storage_client = self._init_client()
        self._parent = parent
        self._bucket_name = bucket_name
        self._bucket = storage_client.bucket(bucket_name)

    def _init_client(self):
        constructor = (storage.Client.from_service_account_json
                       if self._from_json else storage.Client)
        return constructor(**self._client_kwargs)

    def download(self, local, destination=None):
        remote = self._remote_path(local)
        destination = destination or local

        # FIXME: downlod and catch exception to avoid making two API calls
        if self._is_file(remote):
            self._download(destination, remote)
        else:
            counter = 0

            for blob in self._bucket.client.list_blobs(self._bucket_name,
                                                       prefix=remote):
                rel = PurePosixPath(blob.name).relative_to(remote)
                destination_file = Path(destination, *rel.parts)
                destination_file.parent.mkdir(exist_ok=True, parents=True)
                blob.download_to_filename(destination_file)
                counter += 1

            if not counter:
                raise RemoteFileNotFound('Could not download '
                                         f'{local!r} using client {self}: '
                                         'No such file or directory')

    def _is_file(self, remote):
        return self._bucket.blob(remote).exists()

    def _is_dir(self, remote):
        return any(
            self._bucket.client.list_blobs(self._bucket_name, prefix=remote))

    def _download(self, local, remote):
        blob = self._bucket.blob(remote)
        Path(local).parent.mkdir(exist_ok=True, parents=True)
        blob.download_to_filename(local)

    def _upload(self, local):
        remote = self._remote_path(local)
        blob = self._bucket.blob(remote)
        blob.upload_from_filename(local)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_bucket']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        storage_client = self._init_client()
        self._bucket = storage_client.bucket(self._bucket_name)

    def __repr__(self):
        return (f'{type(self).__name__}(bucket_name={self._bucket_name!r}, '
                f'parent={self._parent!r}, '
                f'path_to_project_root={str(self._path_to_project_root)!r})')



"""
Utilities for extracting docstrings
"""
import re


def extract_from_sql(code):
    """Extract from a SQL script
    """
    # [start] [any charaters] [\] [*] [any word chars] [*] [/] [any word chars]
    regex = r'^\s*\/\*([\w\W]+)\*\/[\w\W]*'
    match = re.match(regex, code)
    return '' if match is None else match.group(1)


def extract_from_nb(nb):
    """Extract from a notebook object
    """
    if not nb.cells:
        return ''

    first = nb.cells[0]

    if first.cell_type == 'markdown':
        return first.source
    elif first.cell_type == 'code':
        return extract_from_triple_quotes(first.source)
    else:
        return ''


def extract_from_triple_quotes(code):
    """Extract from code string defined as triple single/double quotes
    """
    return (extract_from_triple_single_quotes(code)
            or extract_from_triple_double_quotes(code))


def extract_from_triple_single_quotes(code):
    regex = r"^\s*'''([\w\W]+)'''[\w\W]*"
    match = re.match(regex, code)
    return '' if match is None else match.group(1)


def extract_from_triple_double_quotes(code):
    regex = r'^\s*"""([\w\W]+)"""[\w\W]*'
    match = re.match(regex, code)
    return '' if match is None else match.group(1)



"""
Products representing SQL relations
"""
import abc
import sqlite3
import json

from jinja2 import Template

from ploomber.products import Product
from ploomber.products.mixins import SQLProductMixin, ProductWithClientMixin
from ploomber.products.serializers import Base64Serializer
from ploomber.placeholders.placeholder import SQLRelationPlaceholder


class SQLiteBackedProductMixin(ProductWithClientMixin, abc.ABC):
    """Mixin for products that store metadata in a SQLite database
    """
    @property
    @abc.abstractmethod
    def name(self):
        """Used as identifier in the database
        """
        pass

    def __create_metadata_relation(self):
        create_metadata = """
        CREATE TABLE IF NOT EXISTS _metadata (
            name TEXT NOT NULL,
            schema TEXT,
            metadata BLOB,
            UNIQUE(schema, name)
        )
        """
        self.client.execute(create_metadata)

    def __get_schema(self):
        # the actual implementation can be a simple identifier or a sql
        # identifier
        if hasattr(self, 'schema'):
            # there is a schema (can be None)
            return self.schema
        else:
            # this means there is no schema
            return False

    def fetch_metadata(self):
        self.__create_metadata_relation()

        query = Template("""
        SELECT metadata FROM _metadata
        WHERE name = '{{name}}'
        {% if schema is not sameas false %}
            {% if schema %}
                AND schema = '{{schema}}'
            {% else %}
                AND schema IS NULL
            {% endif %}
        {% endif %}
        """).render(name=self.name, schema=self.__get_schema())

        cur = self.client.connection.cursor()
        cur.execute(query)
        records = cur.fetchone()
        cur.close()

        if records:
            metadata_bin = records[0]
            return json.loads(metadata_bin.decode("utf-8"))
        else:
            return None

    def save_metadata(self, metadata):
        self.__create_metadata_relation()

        metadata_bin = json.dumps(metadata).encode('utf-8')
        cur = self.client.connection.cursor()
        schema = self.__get_schema()

        if schema is not False:
            if schema is None:
                cur.execute('DELETE FROM _metadata WHERE name = ?',
                            (self.name, ))
            else:
                cur.execute(
                    'DELETE FROM _metadata '
                    'WHERE name = ? AND schema = ?', (self.name, schema))

            # we cannot rely on INSERT INTO since NULL schema values are
            # allowed
            query = """
                INSERT INTO _metadata (metadata, name, schema)
                VALUES(?, ?, ?)
            """
            cur.execute(query,
                        (sqlite3.Binary(metadata_bin), self.name, schema))
        else:
            query = """
                REPLACE INTO _metadata(metadata, name)
                VALUES(?, ?)
            """
            cur.execute(query, (sqlite3.Binary(metadata_bin), self.name))

        self.client.connection.commit()
        cur.close()

    def _delete_metadata(self):
        self.__create_metadata_relation()

        query = Template("""
        DELETE FROM _metadata
        WHERE name = '{{name}}'
        {% if schema is not sameas false %}
            {% if schema %}
                AND schema = '{{schema}}'
            {% else %}
                AND schema IS NULL
            {% endif %}
        {% endif %}
        """).render(name=self.name, schema=self.__get_schema())

        cur = self.client.connection.cursor()
        cur.execute(query)
        cur.close()


class SQLiteRelation(SQLProductMixin, SQLiteBackedProductMixin, Product):
    """A SQLite relation

    Parameters
    ----------
    identifier: tuple of length 3 or 2
        A tuple with (schema, name, kind) where kind must be either 'table'
        or 'view'. If passed a tuple with length 2, schema is assumed None.
        Schemas in SQLite represent other databases when using the ATTACH
        command.
    client: ploomber.clients.DBAPIClient or SQLAlchemyClient, optional
        The client used to connect to the database. Only required
        if no dag-level client has been declared using dag.clients[class]


    Examples
    --------
    >>> from ploomber.products import SQLiteRelation
    >>> relation = SQLiteRelation(('schema', 'some_table', 'table'))
    >>> str(relation) # returns qualified name
    'schema.some_table'
    """
    def __init__(self, identifier, client=None):
        super().__init__(identifier)
        self._client = client

    def _init_identifier(self, identifier):
        return SQLRelationPlaceholder(identifier)

    def exists(self):
        query = """
        SELECT name
        FROM sqlite_master
        WHERE type = '{kind}'
        AND name = '{name}'
        """.format(kind=self.kind, name=self.name)

        cur = self.client.connection.cursor()
        cur.execute(query)
        exists = cur.fetchone() is not None
        cur.close()
        return exists

    def delete(self):
        """Deletes the product
        """
        query = ("DROP {kind} IF EXISTS {relation}".format(kind=self.kind,
                                                           relation=str(self)))
        self.logger.debug(
            'Running "{query}" on the databse...'.format(query=query))
        self.client.execute(query)

    @property
    def name(self):
        return self._identifier.name

    @property
    def schema(self):
        return self._identifier.schema

    @property
    def kind(self):
        return self._identifier.kind

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        return hash((self.schema, self.name, self.kind))

    def __repr__(self):
        return f'{type(self).__name__}({self._identifier._raw_repr()})'


# FIXME: self._identifier should not be accessed direclty since it might
# be a placeholder
class PostgresRelation(SQLProductMixin, ProductWithClientMixin, Product):
    """A PostgreSQL relation

    Parameters
    ----------
    identifier: tuple of length 3
        A tuple with (schema, name, kind) where kind must be either 'table'
        or 'view'
    client: ploomber.clients.DBAPIClient or SQLAlchemyClient, optional
        The client used to connect to the database. Only required
        if no dag-level client has been declared using dag.clients[class]

    Examples
    --------
    >>> from ploomber.products import PostgresRelation
    >>> relation = PostgresRelation(('schema', 'some_table', 'table'))
    >>> str(relation) # returns qualified name
    'schema.some_table'
    """

    # FIXME: identifier has schema as optional but that introduces ambiguity
    # when fetching metadata and checking if the table exists so maybe it
    # should be required

    def __init__(self, identifier, client=None):
        super().__init__(identifier)
        self._client = client

    def _init_identifier(self, identifier):
        return SQLRelationPlaceholder(identifier)

    def fetch_metadata(self):
        cur = self.client.connection.cursor()

        if self.schema:
            schema = self.schema
        else:
            # if schema is empty, we have to find out the default one
            query = """
            SELECT current_schema()
            """
            cur.execute(query)
            schema = cur.fetchone()[0]

        # https://stackoverflow.com/a/11494353/709975
        query = """
        SELECT description
        FROM pg_description
        JOIN pg_class ON pg_description.objoid = pg_class.oid
        JOIN pg_namespace ON pg_class.relnamespace = pg_namespace.oid
        WHERE nspname = %(schema)s
        AND relname = %(name)s
        """
        cur.execute(query, dict(schema=schema, name=self.name))
        metadata = cur.fetchone()

        cur.close()

        # no metadata saved
        if metadata is None:
            return None
        else:
            return Base64Serializer.deserialize(metadata[0])

        # TODO: also check if metadata  does not give any parsing errors,
        # if yes, also return a dict with None values, and maybe emit a warn

    def save_metadata(self, metadata):
        metadata = Base64Serializer.serialize(metadata)

        query = (("COMMENT ON {} {} IS '{}';".format(self.kind,
                                                     self._identifier,
                                                     metadata)))

        cur = self.client.connection.cursor()
        cur.execute(query)
        self.client.connection.commit()
        cur.close()

    def exists(self):
        cur = self.client.connection.cursor()

        if self.schema:
            schema = self.schema
        else:
            # if schema is empty, we have to find out the default one
            query = """
            SELECT current_schema()
            """
            cur.execute(query)
            schema = cur.fetchone()[0]

        # https://stackoverflow.com/a/24089729/709975
        query = """
        SELECT EXISTS (
            SELECT 1
            FROM   pg_catalog.pg_class c
            JOIN   pg_catalog.pg_namespace n ON n.oid = c.relnamespace
            WHERE  n.nspname = %(schema)s
            AND    c.relname = %(name)s
        );
        """

        cur.execute(query, dict(schema=schema, name=self.name))
        exists = cur.fetchone()[0]
        cur.close()
        return exists

    def delete(self, force=False):
        """Deletes the product
        """
        cascade = 'CASCADE' if force else ''
        query = ("DROP {} IF EXISTS {} {}".format(self.kind, self, cascade))
        self.logger.debug('Running "%s" on the databse...', query)

        cur = self.client.connection.cursor()
        cur.execute(query)
        cur.close()
        self.client.connection.commit()

    @property
    def name(self):
        return self._identifier.name

    @property
    def schema(self):
        return self._identifier.schema

    @property
    def kind(self):
        return self._identifier.kind

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        return hash((self.schema, self.name, self.kind))

    def __repr__(self):
        return f'{type(self).__name__}({self._identifier._raw_repr()})'



import copy as copy_module

from ploomber.util import validate
from ploomber.tasks import PythonCallable


def _do_nothing(x):
    return x


class InMemoryDAG:
    """
    Converts a DAG to a DAG-like object that performs all operations in memory
    (products are not serialized). For this to work all tasks must be
    ``PythonCallable`` objects initialized with callables that return a value
    with valid ``serializer`` and ``unserializer`` parameters.

    Parameters
    ----------
    dag : ploomber.DAG
        The DAG to use

    Examples
    --------
    .. literalinclude:: ../../../doc/examples/InMemoryDAG.py
    """
    def __init__(self, dag, return_postprocessor=None):
        types = {type(dag[t]) for t in dag._iter()}

        if not {PythonCallable} >= types:
            extra = ','.join(
                [type_.__name__ for type_ in types - {PythonCallable}])
            raise TypeError('All tasks in the DAG must be PythonCallable, '
                            'got unallowed types: {}'.format(extra))

        # if the tasks are not initialized with a serializer in the partial
        # (this happens when the partial is imported in a pipeline.yaml file
        # and the latter defines DAG-level serializers), the source will
        # be initialized with .source._needs_product=True. We won't pass
        # product here so we force it to False. Othersie dag.render() will fail
        for name in dag._iter():
            dag[name].source._needs_product = False

        dag.render()

        self.dag = dag
        self.root_nodes = [
            name for name, degree in dag._G.in_degree() if not degree
        ]
        self.return_postprocessor = return_postprocessor or _do_nothing

        # TODO: validate that root nodes have single parameter in the signature
        # input_data. and was initialized with {'input_data': None}

        # TODO: maybe raise a warning if the DAG has attributes that won't be
        # used (like serializer, unserializer). Or just document it?

    def build(self, input_data, copy=False):
        """Run the DAG

        Parameters
        ----------
        input_data : dict
            A dictionary mapping root tasks (names) to dict params. Root tasks
            are tasks in the DAG that do not have upstream dependencies,
            the corresponding dictionary is passed to the respective task
            source function as keyword arguments

        copy : bool or callable
            Whether to copy the output of an upstream task before passing it
            to the task being processed. It is recommended to turn this off
            for memory efficiency but if the tasks are not pure functions
            (i.e. mutate their inputs) this migh lead to bugs, in such
            case, the best way to fix it would be to make all your tasks
            pure functions but you can enable this option if memory
            consumption is not a problem. If True it uses the ``copy.copy``
            function before passing the upstream products, if you pass a
            callable instead, such function is used (for example, you
            may pass ``copy.deepcopy``)

        Returns
        -------
        dict
            A dictionary mapping task names to their respective outputs
        """
        outs = {}

        input_data_names = set(self.root_nodes)
        # FIXME: for this particula case, the error here should be TypeError,
        # not KeyError (the former is the one used when calling functions with
        # invalid arguments) - maybe an argument validate.keys to choose
        # which error to raise?
        validate.keys(valid=input_data_names,
                      passed=set(input_data),
                      required=input_data_names,
                      name='input_data')

        if copy is True:
            copying_function = copy_module.copy
        elif callable(copy):
            copying_function = copy
        else:
            copying_function = _do_nothing

        for task_name in self.dag:
            task = self.dag[task_name]
            params = task.params.to_dict()

            if task_name in self.root_nodes:
                params = {**params, 'input_data': input_data[task_name]}

            # replace params with the returned value from upstream tasks
            if 'upstream' in params:
                params['upstream'] = {
                    k: copying_function(outs[k])
                    for k, v in params['upstream'].items()
                }

            params.pop('product', None)

            output = self.return_postprocessor(task.source.primitive(**params))

            if output is None:
                raise ValueError(
                    'All callables in a {} must return a value. '
                    'Callable "{}", from task "{}" returned None'.format(
                        type(self).__name__, task.source.name, task_name))

            outs[task_name] = output

        return outs



from ploomber.static_analysis.python import PythonNotebookExtractor
from ploomber.static_analysis.r import RNotebookExtractor
from ploomber.static_analysis.sql import SQLExtractor

_EXTRACTOR_FOR_LANGUAGE = {
    'python': PythonNotebookExtractor,
    'r': RNotebookExtractor,
    'sql': SQLExtractor
}


def extractor_class_for_language(language):
    if language not in _EXTRACTOR_FOR_LANGUAGE:
        raise NotImplementedError(
            '"{}" is unsupported, supported languages are {}'.format(
                language, list(_EXTRACTOR_FOR_LANGUAGE)))

    return _EXTRACTOR_FOR_LANGUAGE[language]



"""
A placeholder Product that doesn't do anything, used internally for
InMemoryDAG
"""
from ploomber.products import Product
from ploomber.placeholders.placeholder import Placeholder


class EmptyProduct(Product):
    def __init__(self):
        super().__init__(identifier='')

    def _init_identifier(self, identifier):
        return Placeholder(identifier)

    def fetch_metadata(self):
        pass

    def save_metadata(self, metadata):
        pass

    def exists(self):
        pass

    def delete(self, force=False):
        pass



"""
As an open source project, we collect anonymous usage statistics to
prioritize and find product gaps.
This is optional and may be turned off by changing the configuration file:
 inside ~/.ploomber/stats/config.yaml
 Change stats_enabled to False.
See the user stats page for more information:
https://docs.ploomber.io/en/latest/community/user-stats.html

The data we collect is limited to:
1. The Ploomber version currently running.
2. a generated UUID, randomized when the initial install takes place,
    no personal or any identifiable information.
3. Environment variables: OS architecture, Python version etc.
4. Information about the different product phases:
    installation, API calls and errors.

    Relational schema for the telemetry.
    event_id - Unique id for the event
    action - Name of function called i.e. `execute_pipeline_started`
    (see: fn telemetry_wrapper)
    client_time - Client time
    elapsed_time - Total time from start to end of the function call
    pipeline_name_hash - Hash of pipeline name, if any
    python_version - Python version
    num_pipelines - Number of pipelines in repo, if any
    metadata - More information i.e. pipeline success (boolean)
    telemetry_version - Telemetry version

"""
import logging
import datetime
import http.client as httplib
import json
import os
from pathlib import Path
import sys
from uuid import uuid4
from functools import wraps
import platform

import click
import posthog

from ploomber.telemetry import validate_inputs
from ploomber.config import Config
from ploomber import __version__

TELEMETRY_VERSION = '0.3'
DEFAULT_HOME_DIR = '~/.ploomber'
DEFAULT_USER_CONF = 'config.yaml'
DEFAULT_PLOOMBER_CONF = 'uid.yaml'
CONF_DIR = "stats"
posthog.project_api_key = 'phc_P9SpSeypyPwxrMdFn2edOOEooQioF2axppyEeDwtMSP'
PLOOMBER_HOME_DIR = os.getenv("PLOOMBER_HOME_DIR")
# posthog client logs errors which are confusing for users
# https://github.com/PostHog/posthog-python/blob/fd92502d990499a61804034e3feb7e17f64a14a1/posthog/consumer.py#L81
logging.getLogger("posthog").disabled = True


class UserSettings(Config):
    """User-customizable settings
    """
    version_check_enabled: bool = True
    cloud_key: str = None
    user_email: str = None
    stats_enabled: bool = True

    @classmethod
    def path(cls):
        return Path(check_dir_exist(CONF_DIR), DEFAULT_USER_CONF)


class Internal(Config):
    """
    Internal file to store settings (not intended to be modified by the
    user)
    """
    last_version_check: datetime.datetime = None
    uid: str
    first_time: bool = True

    @classmethod
    def path(cls):
        return Path(check_dir_exist(CONF_DIR), DEFAULT_PLOOMBER_CONF)

    def uid_default(self):
        return str(uuid4())


def python_version():
    py_version = sys.version_info
    return f"{py_version.major}.{py_version.minor}.{py_version.micro}"


def is_online():
    """Check if host is online
    """
    conn = httplib.HTTPSConnection('www.google.com', timeout=1)

    try:
        conn.request("HEAD", "/")
        return True
    except Exception:
        return False
    finally:
        conn.close()


def is_docker():
    """Will output if the code is within a container
    """
    try:
        cgroup = Path('/proc/self/cgroup')
        docker_env = Path('/.dockerenv')
        return (docker_env.exists() or cgroup.exists()
                and any('docker' in line
                        for line in cgroup.read_text().splitlines()))
    except OSError:
        return False


def get_os():
    """
    The function will output the client platform
    """
    os = platform.system()
    if os == "Darwin":
        return 'MacOS'
    else:  # Windows/Linux are contained
        return os


def is_conda():
    """
    The function will tell if the code is running in a conda env
    """
    conda_path = Path(sys.prefix, 'conda-meta')
    return conda_path.exists() or os.environ.get(
        "CONDA_PREFIX", False) or os.environ.get("CONDA_DEFAULT_ENV", False)


def get_base_prefix_compat():
    """
    This function will find the pip virtualenv with different python versions.
    Get base/real prefix, or sys.prefix if there is none.
    """
    return getattr(sys, "base_prefix", None) or sys.prefix or getattr(
        sys, "real_prefix", None)


def in_virtualenv():
    return get_base_prefix_compat() != sys.prefix


def get_env():
    """Returns: The name of the virtual env if exists as str"""
    if in_virtualenv():
        return 'pip'
    elif is_conda():
        return 'conda'
    else:
        return 'local'


def is_colab():
    """Returns: True for Google Colab env"""
    return "COLAB_GPU" in os.environ


def is_paperspace():
    """Returns: True for Paperspace env"""
    return "PS_API_KEY" in os.environ or\
           "PAPERSPACE_API_KEY" in os.environ or\
           "PAPERSPACE_NOTEBOOK_REPO_ID" in os.environ


def is_slurm():
    """Returns: True for Slurm env"""
    return "SLURM_JOB_ID" in os.environ


def is_airflow():
    """Returns: True for Airflow env"""
    return "AIRFLOW_CONFIG" in os.environ or "AIRFLOW_HOME" in os.environ


def is_argo():
    """Returns: True for Airflow env"""
    return "ARGO_AGENT_TASK_WORKERS" in os.environ or \
           "ARGO_KUBELET_PORT" in os.environ


def clean_tasks_upstream_products(input):
    clean_input = {}
    try:
        product_items = input.items()
        for product_item_name, product_item in product_items:
            clean_input[product_item_name] = str(product_item).split("/")[-1]
    except AttributeError:  # Single product
        return str(input.split("/")[-1])

    return clean_input


def parse_dag(dag):
    try:
        dag_dict = {}
        dag_dict["dag_size"] = str(len(dag))
        tasks_list = list(dag)
        if tasks_list:
            dag_dict["tasks"] = {}
            for task in tasks_list:
                task_dict = {}
                task_dict["status"] = dag[task]._exec_status.name
                task_dict["type"] = str(type(
                    dag[task])).split(".")[-1].split("'")[0]
                task_dict["upstream"] = clean_tasks_upstream_products(
                    dag[task].upstream)
                task_dict["products"] = clean_tasks_upstream_products(
                    dag[task].product.to_json_serializable())
                dag_dict['tasks'][task] = task_dict

        return dag_dict
    except Exception:
        return None


def get_home_dir():
    """
    Checks if ploomber home was set through the env variable.
    returns the actual home_dir path.
    """
    return PLOOMBER_HOME_DIR if PLOOMBER_HOME_DIR else DEFAULT_HOME_DIR


def check_dir_exist(input_location=None):
    """
    Checks if a specific directory exists, creates if not.
    In case the user didn't set a custom dir, will turn to the default home
    """
    home_dir = get_home_dir()

    if input_location:
        p = Path(home_dir, input_location)
    else:
        p = Path(home_dir)

    p = p.expanduser()

    if not p.exists():
        p.mkdir(parents=True)

    return p


def check_telemetry_enabled():
    """
    Check if the user allows us to use telemetry. In order of precedence:

    1. If PLOOMBER_STATS_ENABLED defined, check its value
    2. Otherwise use the value in stats_enabled in the config.yaml file
    """
    if 'PLOOMBER_STATS_ENABLED' in os.environ:
        return os.environ['PLOOMBER_STATS_ENABLED'].lower() == 'true'

    settings = UserSettings()
    return settings.stats_enabled


def check_first_time_usage():
    """
    The function checks for first time usage if the conf file exists and the
    uid file doesn't exist.
    """
    internal = Internal()
    first_time = internal.first_time
    internal.first_time = False
    return first_time


def get_latest_version():
    """
    The function checks for the latest available ploomber version
    uid file doesn't exist.
    """
    conn = httplib.HTTPSConnection('pypi.org', timeout=1)
    try:
        conn.request("GET", "/pypi/ploomber/json")
        content = conn.getresponse().read()
        data = json.loads(content)
        latest = data['info']['version']
        return latest
    except Exception:
        return __version__
    finally:
        conn.close()


def is_cloud_user():
    """
    The function checks if the cloud api key is set for the user.
    Checks if the cloud_key is set in the User conf file (config.yaml).
    returns True/False accordingly.
    """
    settings = UserSettings()
    return settings.cloud_key


def email_registered():
    """
        The function checks if the email is set for the user.
        Checks if the user_email is set in the User conf file (config.yaml).
        returns True/False accordingly.
        """
    settings = UserSettings()
    return settings.user_email


def check_version():
    """
    The function checks if the user runs the latest version
    This check will be skipped if the version_check_enabled is set to False
    If it's not the latest, notifies the user and saves the metadata to conf
    Alerting every 2 days on stale versions
    """
    settings = UserSettings()

    if not settings.version_check_enabled:
        return

    internal = Internal()
    now = datetime.datetime.now()

    # Check if we already notified in the last 2 days
    if internal.last_version_check and (now -
                                        internal.last_version_check).days < 2:
        return

    # check latest version (this is an expensive call since it hits pypi.org)
    # so we only ping the server when it's been 2 days
    latest = get_latest_version()

    # If latest version, do nothing
    if __version__ == latest:
        return

    click.secho(
        f"There's a new Ploomber version available ({latest}), "
        f"you're running {__version__}. To upgrade: "
        "pip install ploomber --upgrade",
        fg='yellow')

    # Update latest check date
    internal.last_version_check = now


def _get_telemetry_info():
    """
    The function checks for the local config and uid files, returns the right
    values according to the config file (True/False). In addition it checks
    for first time installation.
    """
    # Check if telemetry is enabled, if not skip, else check for uid
    telemetry_enabled = check_telemetry_enabled()

    # Check latest version
    check_version()

    if telemetry_enabled:
        # Check first time install
        is_install = check_first_time_usage()

        # if not uid, create
        internal = Internal()
        return telemetry_enabled, internal.uid, is_install
    else:
        return False, '', False


def validate_entries(event_id, uid, action, client_time, total_runtime):
    event_id = validate_inputs.str_param(str(event_id), "event_id")
    uid = validate_inputs.str_param(uid, "uid")
    action = validate_inputs.str_param(action, "action")
    client_time = validate_inputs.str_param(str(client_time), "client_time")
    elapsed_time = validate_inputs.opt_str_param(str(total_runtime),
                                                 "elapsed_time")
    return event_id, uid, action, client_time, elapsed_time


def log_api(action, client_time=None, total_runtime=None, metadata=None):
    """
    This function logs through an API call, assigns parameters if missing like
    timestamp, event id and stats information.
    """
    metadata = metadata or {}

    event_id = uuid4()

    if client_time is None:
        client_time = datetime.datetime.now()

    (telemetry_enabled, uid, is_install) = _get_telemetry_info()

    # NOTE: this should not happen anymore
    if 'NO_UID' in uid:
        metadata['uid_issue'] = uid
        uid = None

    py_version = python_version()
    docker_container = is_docker()
    cloud = is_cloud_user()
    email = email_registered()
    colab = is_colab()
    if colab:
        metadata['colab'] = colab

    paperspace = is_paperspace()
    if paperspace:
        metadata['paperspace'] = paperspace

    slurm = is_slurm()
    if slurm:
        metadata['slurm'] = slurm

    airflow = is_airflow()
    if airflow:
        metadata['airflow'] = airflow

    argo = is_argo()
    if argo:
        metadata['argo'] = argo

    if 'dag' in metadata:
        metadata['dag'] = parse_dag(metadata['dag'])

    os = get_os()
    product_version = __version__
    online = is_online()
    environment = get_env()

    if telemetry_enabled and online:
        (event_id, uid, action, client_time,
         elapsed_time) = validate_entries(event_id, uid, action, client_time,
                                          total_runtime)
        props = {
            'event_id': event_id,
            'user_id': uid,
            'action': action,
            'client_time': str(client_time),
            'metadata': metadata,
            'total_runtime': total_runtime,
            'python_version': py_version,
            'ploomber_version': product_version,
            'docker_container': docker_container,
            'cloud': cloud,
            'email': email,
            'os': os,
            'environment': environment,
            'metadata': metadata,
            'telemetry_version': TELEMETRY_VERSION
        }

        if is_install:
            posthog.capture(distinct_id=uid,
                            event='install_success_indirect',
                            properties=props)

        posthog.capture(distinct_id=uid, event=action, properties=props)


# NOTE: should we log differently depending on the error type?
# NOTE: how should we handle chained exceptions?
def log_call(action, payload=False):
    """Runs a function and logs it
    """
    def _log_call(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            _payload = dict()
            log_api(action=f'{action}-started', metadata={'argv': sys.argv})
            start = datetime.datetime.now()

            try:
                if payload:
                    result = func(_payload, *args, **kwargs)
                else:
                    result = func(*args, **kwargs)
            except Exception as e:
                log_api(
                    action=f'{action}-error',
                    total_runtime=str(datetime.datetime.now() - start),
                    metadata={
                        # can we log None to posthog?
                        'type': getattr(e, 'type_', None),
                        'exception': str(e),
                        'argv': sys.argv,
                        **_payload
                    })
                raise e
            else:
                log_api(action=f'{action}-success',
                        total_runtime=str(datetime.datetime.now() - start),
                        metadata={
                            'argv': sys.argv,
                            **_payload
                        })

            return result

        return wrapper

    return _log_call



from ploomber.products.product import Product
from ploomber.products.metaproduct import MetaProduct
from ploomber.products.file import File
from ploomber.products.genericproduct import (GenericProduct,
                                              GenericSQLRelation, SQLRelation)
from ploomber.products.sql import SQLiteRelation, PostgresRelation
from ploomber.products.emptyproduct import EmptyProduct

__all__ = [
    'File', 'MetaProduct', 'Product', 'SQLiteRelation', 'PostgresRelation',
    'GenericProduct', 'GenericSQLRelation', 'SQLRelation', 'EmptyProduct'
]



"""
Task abstract class

A Task is a unit of work, it has associated source code and a product
(a persistent object such as a table in a database), it has a name and lives
in a DAG.

[WIP] On subclassing Tasks

Implementation details:

* params (dict), upstream (Param object)
* params vs constructor parameters
* params on render vs params on run
* Implementing Task.run (using the source object, product, TaskBuildError)

Optional:

* Validating PRODUCT_CLASSES_ALLOWED
* Using a client parameter

Task execution model:

Upon instantiation, all tasks start with a WaitingRender status. On successful
render, they move to Skipped if they are up-to-date and all upstream
dependencies are up-to-date as well, WaitingExecution if they
are outdated but upstream dependencies are up-to-date or WaitingUpstream, if
any upstream dependency is outdated.

If render fails, they are marked as ErroredRender and downstream dependencies
are marked as AbortedRender.

When a Task is built, and succeeds, the on_finish hook is executed, if
on_finish succeeds as well, the Task is marked as Executed. If either
the task fails or on_finish fails, the Task is marked as Errored.

When a task is marked as Executed, is saves metadata in its Product,
if the product registered a prepare_metadata hook, the metadata to save
can be modified before is saved.

When a Task is marked as Errored, on_failure hook is executed and downstream
dependencies are marked as Aborted.


Beware that sometimes executors call Task._build in a different process, hence
updates won't be visible in Task objects that exist in the main process.
For that reason, executors have to report back status to the original Task
object using Task.exec_status, see executors documentation for details

TODO: describe BrokenProcesssPool status
"""
import abc
import logging
from datetime import datetime
from collections import defaultdict
from pathlib import Path

from ploomber.products import Product, MetaProduct, EmptyProduct, File
from ploomber.exceptions import (TaskBuildError, DAGBuildEarlyStop,
                                 CallbackCheckAborted)
from ploomber.tasks.taskgroup import TaskGroup
from ploomber.constants import TaskStatus
from ploomber.tasks._upstream import Upstream
from ploomber.tasks._params import Params
from ploomber.tasks.util import download_products_in_parallel
from ploomber.table import TaskReport, Row
from ploomber.util import isiterable
from ploomber.util.util import callback_check
from ploomber.dag.abstractdag import AbstractDAG
from ploomber.io import pretty_print

import humanize


class Task(abc.ABC):
    """
    Abstract class for all Tasks

    Parameters
    ----------
    source: str or pathlib.Path
        Source code for the task, for tasks that do not take source code
        as input (such as PostgresCopyFrom), this can be another thing. The
        source can be a template and can make references to any parameter
        in "params", "upstream" parameters or its own "product", not all
        Tasks have templated source (templating code is mostly used by
        Tasks that take SQL source code as input)
    product: Product
        The product that this task will create upon completion
    dag: DAG
        The DAG holding this task
    name: str
        A name for this task, if None a default will be assigned
    params: dict
        Extra parameters passed to the task on rendering (if templated
        source) or during execution (if not templated source)

    Attributes
    ----------
    params : Params
        A read-only dictionary-like object with params passed, after running
        'product' and 'upstream' are added, if any

    on_render : callable
        Function to execute after rendering. The function can request any of
        the following arguments: task, client, product, and params.

    on_finish : callable
        Function to execute upon execution. Can request the same arguments as
        the on_render hook.

    on_failure : callable
        Function to execute upon failure. Can request the same arguments as
        the on_render hook.

    Notes
    -----
    All subclasses must implement the same constuctor to keep the API
    consistent, optional parameters after "params" are ok
    """
    PRODUCT_CLASSES_ALLOWED = None

    @abc.abstractmethod
    def run(self):
        """This is the only required method Task subclasses must implement
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def _init_source(self, source, kwargs):
        pass  # pragma: no cover

    def __init__(self, product, dag, name=None, params=None):
        self._params = Params(params)

        if name is None:
            # use name inferred from the source object
            self._name = self._source.name

            if self._name is None:
                raise AttributeError('Task name can only be None if it '
                                     'can be inferred from the source object. '
                                     'For example, when the task receives a '
                                     'pathlib.Path, when using SourceLoader '
                                     'or in PythonCallable. Pass a value '
                                     'explicitly')
        else:
            self._name = name

        if not isinstance(dag, AbstractDAG):
            raise TypeError(
                f"'dag' must be an instance of DAG, got {type(dag)!r}")

        # NOTE: we should get rid of this, maybe just add hooks that are
        # called back on the dag object to avoid having a reference here
        self.dag = dag
        dag._add_task(self)

        if not hasattr(self, '_source'):
            raise RuntimeError(
                'self._source must be initialized before calling '
                '__init__ in Task')

        if self._source is None:
            raise TypeError(
                '_init_source must return a source object, got None')

        if isinstance(product, Product):
            self._product = product

            if self.PRODUCT_CLASSES_ALLOWED is not None:
                if not isinstance(self._product, self.PRODUCT_CLASSES_ALLOWED):
                    raise TypeError('{} only supports the following product '
                                    'classes: {}, got {}'.format(
                                        type(self).__name__,
                                        self.PRODUCT_CLASSES_ALLOWED,
                                        type(self._product).__name__))
        else:
            # if assigned a tuple/list of products, create a MetaProduct
            self._product = MetaProduct(product)

            if self.PRODUCT_CLASSES_ALLOWED is not None:
                if not all(
                        isinstance(p, self.PRODUCT_CLASSES_ALLOWED)
                        for p in self._product):
                    raise TypeError('{} only supports the following product '
                                    'classes: {}, got {}'.format(
                                        type(self).__name__,
                                        self.PRODUCT_CLASSES_ALLOWED,
                                        type(self._product).__name__))

        self._logger = logging.getLogger('{}.{}'.format(
            __name__,
            type(self).__name__))

        self.product.task = self
        self._client = None

        self.exec_status = TaskStatus.WaitingRender

        self._on_finish = None
        self._on_failure = None
        self._on_render = None

    @property
    def _available_callback_kwargs(self):
        # make it a property so we always get the latest value for self.client
        # given that could be None during init
        return {
            'task': self,
            'client': self.client,
            'product': self.product,
            'params': self.params
        }

    @property
    def client(self):
        return self._client

    @property
    def name(self):
        """
        A str that represents the name of the task, you can access tasks
        in a dag using dag['some_name']
        """
        return self._name

    @property
    def source(self):
        """
        Source is used by the task to compute its output, for most cases
        this is source code, for example PythonCallable takes a function
        as source and SQLScript takes a string with SQL code as source.
        But other tasks might take non-code objects as source, for example,
        PostgresCopyFrom takes a path to a file. If source represents code
        doing str(task.source) will return the string representation
        """
        return self._source

    @property
    def product(self):
        """The product this task will create upon execution
        """
        return self._product

    @property
    def upstream(self):
        """
        A mapping for upstream dependencies {task name} -> [task object]
        """
        # this is just syntactic sugar, upstream relations are tracked by the
        # DAG object

        # this always return a copy to prevent global state if contents
        # are modified (e.g. by using pop)
        return self.dag._get_upstream(self.name)

    @property
    def _upstream_product_grouped(self):
        """
        Similar to .upstream but this one nests groups. Output will be the
        same as .upstream if no upstream dependencies are grouped. This is
        only used internally in .render to correctly pass upstream to
        task.params.

        Unlike .upstream, this method returns products instead of task names
        """
        grouped = defaultdict(lambda: {})

        for up_name, up_task in self.upstream.items():
            data = self.dag._G.get_edge_data(up_name, self.name)

            if data and 'group_name' in data:
                group_name = data['group_name']
                grouped[group_name][up_name] = up_task.product
            else:
                grouped[up_name] = up_task.product

        return dict(grouped)

    @property
    def params(self):
        """
        dict that holds the parameter that will be passed to the task upon
        execution. Before rendering, this will only hold parameters passed
        in the Task constructor. After rendering, this will hold new keys:
        "product" contained the rendered product and "upstream" holding
        upstream parameters if there is any
        """
        return self._params

    @property
    def _lineage(self):
        """
        Set with task names of all the dependencies for this task
        (including dependencies of dependencies)
        """
        # if no upstream deps, there is no lineage
        if not len(self.upstream):
            return None
        else:
            # retrieve lineage: upstream tasks + lineage from upstream tasks
            up = list(self.upstream.keys())
            lineage_up = [
                up._lineage for up in self.upstream.values() if up._lineage
            ]
            lineage = up + [task for lineage in lineage_up for task in lineage]
            return set(lineage)

    @property
    def on_finish(self):
        """
        Callable to be executed after this task is built successfully
        (passes Task as first parameter)
        """
        return self._on_finish

    @on_finish.setter
    def on_finish(self, value):
        try:
            callback_check(value, self._available_callback_kwargs)
        # raised when value is a dotted path with lazy loading turned on.
        # cannot check because that requires importing the dotted path
        except CallbackCheckAborted:
            pass

        self._on_finish = value

    def _run_on_finish(self):
        """Call on_finish hook
        """
        if self.on_finish:
            kwargs = callback_check(self.on_finish,
                                    self._available_callback_kwargs)
            self.on_finish(**kwargs)

    def _post_run_actions(self):
        """
        Call on_finish hook, save metadata, verify products exist and upload
        product
        """
        # run on finish first, if this fails, we don't want to save metadata
        try:
            self._run_on_finish()
        except Exception:
            # NOTE: we also set the status in Task._build, which runs during
            # DAG.build() - but setting if here as well to prevent DAG
            # inconsistent state when the user calls Tas.build() directly
            self.exec_status = TaskStatus.Errored
            raise

        if self.exec_status == TaskStatus.WaitingDownload:
            # clear current metadata to force reload
            # and ensure the task uses the downloaded metadata
            self.product.metadata.clear()
        else:
            self.product.metadata.update(
                source_code=str(self.source),
                params=self.params.to_json_serializable(params_only=True))

        # For most Products, it's ok to do this check before
        # saving metadata, but not for GenericProduct, since the way
        # exists() works is by checking metadata, so we have to do it
        # here, after saving metadata
        if isinstance(self.product, MetaProduct):
            missing = self.product.missing()

            if missing:
                raise TaskBuildError(
                    f'Error building task {self.name!r}: '
                    'the task ran successfully but the following '
                    'products are '
                    f'missing: {pretty_print.iterable(missing, repr_=True)}')

        elif not self.product.exists():
            raise TaskBuildError(f'Error building task {self.name!r}: '
                                 'the task ran successfully but the '
                                 f'product {self.product!r} is missing. '
                                 'Ensure your code is generating it.')

        if self.exec_status != TaskStatus.WaitingDownload:
            self.product.upload()

    @property
    def on_failure(self):
        """
        Callable to be executed if task fails (passes Task as first parameter
        and the exception as second parameter)
        """
        return self._on_failure

    @on_failure.setter
    def on_failure(self, value):
        try:
            callback_check(value, self._available_callback_kwargs)
        # raised when value is a dotted path with lazy loading turned on.
        # cannot check because that requires importing the dotted path
        except CallbackCheckAborted:
            pass
        self._on_failure = value

    def _run_on_failure(self):
        if self.on_failure:
            kwargs = callback_check(self.on_failure,
                                    self._available_callback_kwargs)
            self.on_failure(**kwargs)

    @property
    def on_render(self):
        return self._on_render

    @on_render.setter
    def on_render(self, value):
        try:
            callback_check(value, self._available_callback_kwargs)
        # raised when value is a dotted path with lazy loading turned on.
        # cannot check because that requires importing the dotted path
        except CallbackCheckAborted:
            pass
        self._on_render = value

    def _run_on_render(self):
        if self.on_render:
            self._logger.debug('Calling on_render hook on task %s', self.name)

            kwargs = callback_check(self.on_render,
                                    self._available_callback_kwargs)
            try:
                self.on_render(**kwargs)
            except Exception as e:
                msg = ('Exception when running on_render '
                       f'for task {self.name!r}')
                self._logger.exception(msg)
                self.exec_status = TaskStatus.ErroredRender
                raise type(e)(msg) from e

    @property
    def exec_status(self):
        return self._exec_status

    @exec_status.setter
    def exec_status(self, value):
        # FIXME: this should only be used for th eexecutor to report back
        # status Executed or Errored, reject all other cases, those are handled
        # internally
        if value not in list(TaskStatus):
            raise ValueError(
                'Setting task.exec_status to an unknown '
                'value: %s', value)

        self._logger.debug('Setting "%s" status to %s', self.name, value)
        self._exec_status = value

        # process might crash, propagate now or changes might not be
        # reflected (e.g. if a Task is marked as Aborted, all downtream
        # tasks should be marked as aborted as well)
        # FIXME: this is inefficient, it is better to traverse
        # the dag in topological order but exclude nodes not affected by
        # this change
        self._update_downstream_status()

    def build(self, force=False, catch_exceptions=True):
        """Build a single task

        Although Tasks are primarily designed to execute via DAG.build(), it
        is possible to do so in isolation. However, this only works if the
        task does not have any unrendered upstream dependencies, if that's the
        case, you should call DAG.render() before calling Task.build()

        Examples
        --------
        >>> from pathlib import Path
        >>> from ploomber import DAG
        >>> from ploomber.tasks import PythonCallable
        >>> from ploomber.products import File
        >>> def fn(product):
        ...     Path(str(product)).touch()
        >>> task = PythonCallable(fn, File('file.txt'), dag=DAG())
        >>> task.build() # doctest: +SKIP

        Returns
        -------
        dict
            A dictionary with keys 'run' and 'elapsed'

        Raises
        ------
        TaskBuildError
            If the error failed to build because it has upstream dependencies,
            the build itself failed or build succeded but on_finish hook failed

        DAGBuildEarlyStop
            If any task or on_finish hook raises a DAGBuildEarlyStop error
        """
        # This is the public API for users who'd to run tasks in isolation,
        # we have to make sure we clear product cache status, otherwise
        # this will interfere with other render calls
        self.render(force=force)

        upstream_exec_status = [t.exec_status for t in self.upstream.values()]

        if any(exec_status == TaskStatus.WaitingRender
               for exec_status in upstream_exec_status):
            raise TaskBuildError('Cannot directly build task "{}" as it '
                                 'has upstream dependencies, call '
                                 'dag.render() first'.format(self.name))

        # we can execute an individual tasks if missing up-to-date upstream
        # dependencies exist in remote storage
        if self.exec_status == TaskStatus.WaitingUpstream:
            ok = {
                t
                for t in self.upstream.values() if t.exec_status in
                {TaskStatus.Skipped, TaskStatus.WaitingDownload}
            }

            not_ok = set(self.upstream.values()) - ok

            if not_ok:
                raise TaskBuildError(
                    f'Cannot build task {self.name!r} because '
                    'the following upstream dependencies are '
                    f'missing: {[t.name for t in not_ok]!r}. Execute upstream '
                    'tasks first. If upstream tasks generate File(s) and you '
                    'configured a File.client, you may also upload '
                    'up-to-date copies to remote storage and they will be '
                    'automatically downloaded')

            download_products_in_parallel(
                t for t in ok if t.exec_status == TaskStatus.WaitingDownload)

        # at this point the task must be WaitingDownload or WaitingExecution
        res, _ = self._build(catch_exceptions=catch_exceptions)

        self.product.metadata.clear()
        return res

    def _build(self, catch_exceptions):
        """
        Private API for building DAGs. This is what executors should call.
        Unlike the public method, this one does not call render, as it
        should happen via a dag.render() call. It takes care of running the
        task and updating status accordingly

        Parameters
        ----------
        catch_exceptions : bool
            If True, catches exceptions during execution and shows a chained
            exception at the end: [original exception] then
            [exception with context info]. Set it to False when debugging
            tasks to drop-in a debugging session at the failing line.
        """
        if not catch_exceptions:
            res = self._run()
            self._post_run_actions()
            return res, self.product.metadata.to_dict()
        else:
            try:
                # TODO: this calls download, if this happens. should
                # hooks be executed when dwnloading? if so, we could
                # change the ran? column from the task report to something
                # like:
                # ran/downloaded/skipped and use that to determine if we should
                # run hooks
                res = self._run()
            except Exception as e:
                msg = 'Error building task "{}"'.format(self.name)
                self._logger.exception(msg)
                self.exec_status = TaskStatus.Errored

                # if there isn't anything left to run, raise exception here
                if self.on_failure is None:
                    if isinstance(e, DAGBuildEarlyStop):
                        raise DAGBuildEarlyStop(
                            'Stopping task {} gracefully'.format(
                                self.name)) from e
                    else:
                        # FIXME: this makes the traceback longer, consider
                        # removing it. The only information this nested
                        # exception provides is the name of the task but we
                        # are still able to provide that if theh executor
                        # has the option to capture exceptions turned on.
                        # An option to consider is to
                        raise TaskBuildError(msg) from e

                build_success = False
                build_exception = e
            else:
                build_success = True
                build_exception = None

            if build_success:
                try:
                    self._post_run_actions()
                except Exception as e:
                    self.exec_status = TaskStatus.Errored
                    msg = ('Exception when running on_finish '
                           f'for task {self.name!r}')
                    self._logger.exception(msg)

                    if isinstance(e, DAGBuildEarlyStop):
                        raise DAGBuildEarlyStop(
                            'Stopping task {} gracefully'.format(
                                self.name)) from e
                    else:
                        raise TaskBuildError(msg) from e
                else:
                    # sucessful task execution, on_finish hook execution,
                    # metadata saving and upload
                    self.exec_status = TaskStatus.Executed

                return res, self.product.metadata.to_dict()
            # error bulding task
            else:
                try:
                    self._run_on_failure()
                except Exception as e:
                    msg = ('Exception when running on_failure '
                           f'for task {self.name!r}')
                    self._logger.exception(msg)
                    raise TaskBuildError(msg) from e

                if isinstance(build_exception, DAGBuildEarlyStop):
                    raise DAGBuildEarlyStop(
                        'Stopping task {} gracefully'.format(
                            self.name)) from build_exception
                else:
                    msg = 'Error building task "{}"'.format(self.name)
                    raise TaskBuildError(msg) from build_exception

    def _run(self):
        """
        Run or download task if certain status conditions are met, otherwise
        raise a TaskBuildError exception
        """
        # cannot keep running, we depend on the render step to get all the
        # parameters resolved (params, upstream, product)
        if self.exec_status == TaskStatus.WaitingRender:
            raise TaskBuildError('Error building task "{}". '
                                 'Cannot build task that has not been '
                                 'rendered, call DAG.render() first'.format(
                                     self.name))

        elif self.exec_status == TaskStatus.Aborted:
            raise TaskBuildError('Attempted to run task "{}", whose '
                                 'status is TaskStatus.Aborted'.format(
                                     self.name))
        elif self.exec_status == TaskStatus.Skipped:
            raise TaskBuildError('Attempted to run task "{}", whose '
                                 'status TaskStatus.Skipped. Render again and '
                                 'set force=True if you want to force '
                                 'execution'.format(self.name))

        # NOTE: should i fetch metadata here? I need to make sure I have
        # the latest before building
        self._logger.info('Starting execution: %s', repr(self))

        then = datetime.now()

        _ensure_parents_exist(self.product)

        if self.exec_status == TaskStatus.WaitingDownload:
            try:
                self.product.download()
            except Exception as e:
                raise TaskBuildError(
                    f'Error downloading Product {self.product!r} '
                    f'from task {self!r}. Check the full traceback above for '
                    'details') from e

        # NOTE: should we validate status here?
        # (i.e., check it's WaitingExecution)
        else:
            self.run()

        now = datetime.now()

        elapsed = (now - then).total_seconds()
        self._logger.info(
            'Done. Operation took {:.1f} seconds'.format(elapsed))

        # TODO: also check that the Products were updated:
        # if they did not exist, they must exist now, if they alredy
        # exist, timestamp must be recent equal to the datetime.now()
        # used. maybe run fetch metadata again and validate?

        return TaskReport.with_data(name=self.name, ran=True, elapsed=elapsed)

    def render(self, force=False, outdated_by_code=True, remote=False):
        """
        Renders code and product, all upstream tasks must have been rendered
        first, for that reason, this method will usually not be called
        directly but via DAG.render(), which renders in the right order.

        Render fully determines whether a task should run or not.

        Parameters
        ----------
        force : bool, default=False
            If True, mark status as WaitingExecution/WaitingUpstream even if
            the task is up-to-date (if there are any File(s) with clients, this
            also ignores the status of the remote copy), otherwise, the normal
            process follows and only up-to-date tasks are marked as Skipped.

        outdated_by_code : bool, default=True
            Factors to determine if Task.product is marked outdated when source
            code changes. Otherwise just the upstream timestamps are used.

        remote : bool, default=False
            Use remote metadata to determine status

        Notes
        -----
        This method tries to avoid calls to check for product status whenever
        possible, since checking product's metadata can be a slow operation
        (e.g. if metadata is stored in a remote database)

        When passing force=True, product's status checking is skipped
        altogether, this can be useful when we only want to quickly get
        a rendered DAG object to interact with it
        """
        self._logger.debug('Calling render on task %s', self.name)

        try:
            self._render_product()
        except Exception as e:
            self.exec_status = TaskStatus.ErroredRender
            raise type(e)('Error rendering product from Task "{}", '
                          ' check the full traceback above for details. '
                          'Task params: {}'.format(repr(self),
                                                   self.params)) from e

        # product does not becomes part of the task parameters when passing
        # an EmptyProduct - this special kind of task is used by InMemoryDAG.
        if not isinstance(self.product, EmptyProduct):
            self.params._setitem('product', self.product)

        try:
            self.source.render(self.params)
        except Exception as e:
            self.exec_status = TaskStatus.ErroredRender
            raise e

        # we use outdated status several time, this caches the result
        # for performance reasons. TODO: move this to Product and make it
        # a property
        is_outdated = ProductEvaluator(self.product,
                                       outdated_by_code,
                                       remote=remote)

        # task with no dependencies
        if not self.upstream:
            # nothing to do, just mark it ready for execution
            if force:
                self._exec_status = TaskStatus.WaitingExecution
                self._logger.debug(
                    'Forcing status "%s", outdated conditions'
                    ' ignored...', self.name)

            # task is outdated, check if we need to execute or download
            elif is_outdated.check():
                # This only happens with File
                if is_outdated.check() == TaskStatus.WaitingDownload:
                    self._exec_status = TaskStatus.WaitingDownload
                else:
                    self._exec_status = TaskStatus.WaitingExecution

            # task is up-to-date
            else:
                self._exec_status = TaskStatus.Skipped

        # tasks with dependencies
        else:
            upstream_exec_status = set(t.exec_status
                                       for t in self.upstream.values())

            all_upstream_ready = upstream_exec_status <= {
                TaskStatus.Executed, TaskStatus.Skipped
            }

            # some upstream tasks need execution (or download)
            if not all_upstream_ready:
                if force or is_outdated.check() is True:
                    self._exec_status = TaskStatus.WaitingUpstream
                elif is_outdated.check() == TaskStatus.WaitingDownload:
                    self._exec_status = TaskStatus.WaitingDownload
                else:
                    self._exec_status = TaskStatus.Skipped

            # all upstream ready
            else:
                if force or is_outdated.check() is True:
                    self._exec_status = TaskStatus.WaitingExecution
                elif is_outdated.check() == TaskStatus.WaitingDownload:
                    self._exec_status = TaskStatus.WaitingDownload
                else:
                    self._exec_status = TaskStatus.Skipped

        self._run_on_render()

    def set_upstream(self, other, group_name=None):
        self.dag._add_edge(other, self, group_name=group_name)

    def status(self, return_code_diff=False, sections=None):
        """Prints the current task status

        Parameters
        ----------
        sections : list, optional
            Sections to include. Defaults to "name", "last_run",
            "oudated", "product", "doc", "location"
        """
        sections = sections or [
            'name', 'last_run', 'outdated', 'product', 'doc', 'location'
        ]

        p = self.product

        data = {}

        if 'name' in sections:
            data['name'] = self.name

        if 'type' in sections:
            data['type'] = type(self).__name__

        if 'status' in sections:
            data['status'] = self.exec_status.name

        if 'client' in sections:
            # FIXME: all tasks should have a client property
            data['client'] = (repr(self.client)
                              if hasattr(self, 'client') else None)

        if 'last_run' in sections:
            if p.metadata.timestamp is not None:
                dt = datetime.fromtimestamp(p.metadata.timestamp)
                date_h = dt.strftime('%b %d, %Y at %H:%M')
                time_h = humanize.naturaltime(dt)
                data['Last run'] = '{} ({})'.format(time_h, date_h)
            else:
                data['Last run'] = 'Has not been run'

        outd_data = p._outdated_data_dependencies()
        outd_code = p._outdated_code_dependency()

        outd = False

        if outd_code:
            outd = 'Source code'

        if outd_data:
            if not outd:
                outd = 'Upstream'
            else:
                outd += ' & Upstream'

        if 'outdated' in sections:
            data['Outdated?'] = outd

        if 'outdated_dependencies' in sections:
            data['Outdated dependencies'] = outd_data

        if 'outdated_code' in sections:
            data['Outdated code'] = outd_code

        if outd_code and return_code_diff:
            data['Code diff'] = (self.dag.differ.get_diff(
                p.metadata.stored_source_code,
                str(self.source),
                extension=self.source.extension))
        else:
            outd_code = ''

        if 'product_type' in sections:
            data['Product type'] = type(self.product).__name__

        if 'product' in sections:
            data['Product'] = repr(self.product)

        if 'product_client' in sections:
            # FIXME: all products should have a client property
            data['Product client'] = (repr(self.product.client) if hasattr(
                self.product, 'client') else None)

        if 'doc' in sections:
            data['Doc (short)'] = _doc_short(self.source.doc)

        if 'location' in sections:
            data['Location'] = self.source.loc

        return Row(data)

    def debug(self):
        """Debug task, only implemented in certain tasks
        """
        raise NotImplementedError(
            '"debug" is not implemented in "{}" tasks'.format(
                type(self).__name__))

    def develop(self):
        """Develop task, only implemented in certain tasks
        """
        raise NotImplementedError(
            '"develop" is not implemented in "{}" tasks'.format(
                type(self).__name__))

    def load(self):
        """Load task as pandas.DataFrame. Only implemented in certain tasks
        """
        raise NotImplementedError(
            '"load" is not implemented in "{}" tasks'.format(
                type(self).__name__))

    def _render_product(self):
        params_names = list(self.params)

        # add upstream product identifiers to params, if any
        # Params are read-only for users, but we have to add upstream
        # dependencies so we do it directly to the dictionary
        # TODO: process wilcards such as fit-*
        if self.upstream:
            self.params._setitem(
                'upstream',
                Upstream(self._upstream_product_grouped, name=self.name))

        # render the current product
        try:
            # using the upstream products to define the current product
            # is optional, using the parameters passed in params is also
            # optional
            self.product.render(self.params,
                                optional=set(params_names + ['upstream']))
        except Exception as e:
            raise type(e)('Error rendering Product from Task "{}", '
                          ' check the full traceback above for details'.format(
                              repr(self))) from e

    def _get_downstream(self):
        # make the _get_downstream more efficient by
        # using the networkx data structure directly
        downstream = []
        for t in self.dag.values():
            if self in t.upstream.values():
                downstream.append(t)
        return downstream

    def _update_downstream_status(self):
        # FIXME: this is inefficient, it is better to traverse
        # the dag in topological order but exclude nodes not affected by
        # this change
        # TODO: move to DAG
        def update_status(task):
            any_upstream_errored_or_aborted = any([
                t.exec_status in (TaskStatus.Errored, TaskStatus.Aborted)
                for t in task.upstream.values()
            ])
            all_upstream_ready = all([
                t.exec_status in {TaskStatus.Executed, TaskStatus.Skipped}
                for t in task.upstream.values()
            ])

            if any_upstream_errored_or_aborted:
                task.exec_status = TaskStatus.Aborted
            elif any([
                    t.exec_status
                    in (TaskStatus.ErroredRender, TaskStatus.AbortedRender)
                    for t in task.upstream.values()
            ]):
                task.exec_status = TaskStatus.AbortedRender
            # mark as waiting execution if moving from waiting upstream
            # the second situation if when task is waitingdownload, in which
            # case we don't do anything
            elif all_upstream_ready and (task.exec_status
                                         == TaskStatus.WaitingUpstream):
                task.exec_status = TaskStatus.WaitingExecution

        for t in self._get_downstream():
            update_status(t)

    def __rshift__(self, other):
        """ a >> b is the same as b.set_upstream(a)
        """
        other.set_upstream(self)
        # return other so a >> b >> c works
        return other

    def __add__(self, other):
        """ a + b means TaskGroup([a, b])
        """
        if isiterable(other) and not isinstance(other, AbstractDAG):
            return TaskGroup([self] + list(other))
        else:
            return TaskGroup((self, other))

    def __repr__(self):
        return ('{}: {} -> {}'.format(
            type(self).__name__, self.name, repr(self.product)))

    def __str__(self):
        return str(self.product)

    def __getstate__(self):
        state = self.__dict__.copy()
        # _logger is not pickable, so we remove them and build
        # them again in __setstate__
        del state['_logger']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._logger = logging.getLogger('{}.{}'.format(
            __name__,
            type(self).__name__))


def _doc_short(doc):
    if doc is not None:
        return doc.split('\n')[0]
    else:
        return None


class ProductEvaluator:
    """
    A class to temporarily keep the outdated status of a product (when products
    are remote this operation is expensive)
    """
    def __init__(self, product, outdated_by_code, remote):
        self.product = product
        self.outdated_by_code = outdated_by_code
        self._is_outdated = None
        self._remote = remote

    def check(self):
        if self._is_outdated is None:
            callable_ = (self.product._is_remote_outdated
                         if self._remote else self.product._is_outdated)

            self._is_outdated = (callable_(
                outdated_by_code=self.outdated_by_code))

        return self._is_outdated


def _ensure_parents_exist(product):
    if isinstance(product, MetaProduct):
        for prod in product:
            _ensure_file_parents_exist(prod)
    else:
        _ensure_file_parents_exist(product)


def _ensure_file_parents_exist(product):
    if isinstance(product, File):
        Path(product).parent.mkdir(parents=True, exist_ok=True)



import shutil
import os
import re
import ast
import pydoc
import getpass
from copy import deepcopy, copy
from collections.abc import Mapping, Iterable
from pathlib import Path
from functools import reduce
import datetime

from jinja2 import Template

from ploomber.placeholders import util
from ploomber import repo
from ploomber.util import default
from ploomber.exceptions import BaseException


def expand_raw_dictionary_and_extract_tags(raw, mapping):
    data = deepcopy(raw)

    placeholders_all = []

    for (d, current_key, current_val, _) in iterate_nested_dict(data):
        d[current_key], placeholders = expand_if_needed(current_val, mapping)
        placeholders_all.extend(placeholders)

    return data, set(placeholders_all)


def expand_raw_dictionary(raw, mapping):
    """
    Expands a dictionary where some values are {{tags}} using their values
    in a mapping
    """
    return expand_raw_dictionary_and_extract_tags(raw, mapping)[0]


def expand_raw_dictionaries_and_extract_tags(raw, mapping):
    """
    Expands a list of dictionaries
    """
    expanded, tags = list(
        zip(*[
            expand_raw_dictionary_and_extract_tags(element, mapping)
            for element in raw
        ]))

    tags_unique = set(reduce(lambda x, y: x | y, tags))

    return expanded, tags_unique


def expand_if_needed(raw_value, mapping):
    value, placeholders = mapping._render(raw_value)
    return cast_if_possible(value), placeholders


def cast_if_possible(value):
    """
    Reference to env in specs must be strings, but we would like the rendered
    value to still have the appropriate type
    """
    if isinstance(value, str):
        value_lower = value.lower()

        if value_lower == 'false':
            return False
        elif value_lower == 'true':
            return True
        elif value_lower in {'none', 'null'}:
            return None

        try:
            return ast.literal_eval(value)
        except Exception:
            pass

    return value


class EnvironmentExpander:
    """
    Convert values in the raw dictionary by expanding tags such as {{git}},
    {{version}} or {{here}}. See `expand_raw_value` for more details

    Parameters
    ----------
    preprocessed : dict
        Preprocessed env dictionary

    path_to_here : str
        Path to env.yaml, used to expand {{here}}

    version_requires_import : bool, default=False
        Whether determining package version requires import or not. If False,
        the root ``__init__.py`` file in the module must have
        a ``__version__ = 'LITERAL'`` variable. Literal is extracted.

    path_to_env : str, default=None
        Path to the env.yaml. If not None, it is used to hint the user which
        file to fix if there are errors.
    """
    def __init__(self,
                 preprocessed,
                 path_to_here=None,
                 version_requires_import=False,
                 path_to_env=None):
        self._preprocessed = preprocessed
        self._path_to_env = path_to_env

        # {{here}} resolves to this value
        self._path_to_here = (None if path_to_here is None else str(
            Path(path_to_here).resolve()))
        # we compute every placeholder's value so we only do it once
        self._placeholders = {}

        self._version_requires_import = version_requires_import

    def expand_raw_dictionary(self, raw):
        data = deepcopy(raw)

        for (d, current_key, current_val,
             parent_keys) in iterate_nested_dict(data):
            d[current_key] = self.expand_raw_value(current_val, parent_keys)

        return data

    def expand_raw_value(self, raw_value, parents):
        """
        Expand a string with placeholders

        Parameters
        ----------
        raw_value : str
            The original value to expand
        parents : list
            The list of parents to get to this value in the dictionary

        Notes
        -----
        If for a given raw_value, the first parent is 'path', expanded value
        is casted to pathlib.Path object and .expanduser() is called,
        furthermore, if raw_value ends with '/', a directory is created if
        it does not currently exist
        """
        placeholders = util.get_tags_in_str(raw_value)

        if not placeholders:
            value = raw_value
        else:
            if 'git' in placeholders:
                if not shutil.which('git'):
                    raise BaseException('Found placeholder {{git}}, but '
                                        'git is not installed. Please install '
                                        'it and try again.')

                if not repo.is_repo(
                        self._preprocessed.get('_module', self._path_to_here)):
                    raise BaseException(
                        'Found placeholder {{git}}, but could not '
                        'locate a git repository. Create a repository '
                        'or remove the {{git}} placeholder.')

            # get all required placeholders
            params = {k: self.load_placeholder(k) for k in placeholders}
            value = Template(raw_value).render(**params)

        if parents:
            if parents[0] == 'path':

                # value is a str (since it was loaded from a yaml file),
                # if it has an explicit trailing slash, interpret it as
                # a directory and create it, we have to do it at this point,
                # because once we cast to Path, we lose the trailing slash
                if str(value).endswith('/'):
                    self._try_create_dir(value)

                return Path(value).expanduser()
            else:
                return cast_if_possible(value)

    def _try_create_dir(self, value):
        # make sure to expand user to avoid creating a "~" folder
        path = Path(value).expanduser()

        if not path.exists():
            path.mkdir(parents=True)

    def load_placeholder(self, key):
        if key not in self._placeholders:
            if hasattr(self, 'get_' + key):
                try:
                    value = getattr(self, 'get_' + key)()
                except Exception as e:
                    raise BaseException('An error happened while '
                                        'expanding placeholder {{' + key +
                                        '}}') from e

                self._placeholders[key] = value
            else:
                if self._path_to_env:
                    msg = ('Error resolving env '
                           f'at {str(self._path_to_env)!r}: '
                           f'Undeclared value for placeholder {key!r}')
                else:
                    msg = ('Error resolving env: Undeclared '
                           f'value for placeholder {key!r}')

                raise BaseException(msg)

        return self._placeholders[key]

    def _get_version_importing(self):
        module_path = self._preprocessed.get('_module')

        if not module_path:
            raise KeyError('_module key is required to use version '
                           'placeholder')

        # is this ok to do? /path/to/{module_name}
        module_name = str(Path(module_path).name)
        module = pydoc.locate(module_name)

        if module is None:
            raise ImportError(
                'Unabe to import module with name "{}"'.format(module_name))

        if hasattr(module, '__version__'):
            return module.__version__
        else:
            raise RuntimeError('Module "{}" does not have a __version__ '
                               'attribute '.format(module))

    def _get_version_without_importing(self):
        if '_module' not in self._preprocessed:
            raise KeyError('_module key is required to use version '
                           'placeholder')

        content = (self._preprocessed['_module'] / '__init__.py').read_text()

        version_re = re.compile(r'__version__\s+=\s+(.*)')

        version = str(ast.literal_eval(version_re.search(content).group(1)))
        return version

    def get_version(self):
        if self._version_requires_import:
            return self._get_version_importing()
        else:
            return self._get_version_without_importing()

    def get_user(self):
        return getpass.getuser()

    def get_cwd(self):
        return str(Path(os.getcwd()).resolve())

    def get_root(self):
        root = default.try_to_find_root_recursively(
            starting_dir=self._path_to_here)

        if root is None:
            raise ValueError('Failed to expand {{root}}, could not '
                             'find a setup.py in a parent folder')

        return root

    def get_here(self):
        if self._path_to_here:
            return self._path_to_here
        else:
            raise RuntimeError('here placeholder is only available '
                               'when env was initialized from a file or '
                               'when directly passing path to use')

    def get_git(self):
        module_path = self._preprocessed.get('_module')

        if self._path_to_here:
            module_path = self._path_to_here

        if not module_path:
            raise KeyError('_module key is required to use git placeholder')

        return repo.git_location(module_path)

    def get_git_hash(self):
        module_path = self._preprocessed.get('_module')

        if self._path_to_here:
            module_path = self._path_to_here

        if not module_path:
            raise KeyError(
                '_module key is required to use git_hash placeholder')

        return repo.git_hash(module_path)

    def get_now(self):
        """Returns current timestamp in ISO 8601 format
        """
        return datetime.datetime.now().isoformat()


def iterate_nested_dict(d):
    """
    Iterate over all values (possibly nested) in a dictionary

    Yields: dict holding the value, current key, current value, list of keys
    to get to this value
    """
    for k, v in d.items():
        for i in _iterate(d, k, v, preffix=[k]):
            yield i


def _iterate(parent, key, value, preffix):
    if isinstance(value, Mapping):
        for k, v in value.items():
            preffix_new = copy(preffix)
            preffix_new.append(k)
            for i in _iterate(value, k, v, preffix_new):
                yield i
    elif isinstance(value, Iterable) and not isinstance(value, str):
        for idx, some_val in enumerate(value):
            preffix_new = copy(preffix)
            preffix_new.append(idx)
            for i in _iterate(value, idx, some_val, preffix_new):
                yield i
    else:
        yield parent, key, value, preffix



import json

from ploomber.constants import TaskStatus
from ploomber.products.metadata import Metadata


class _RemoteFile:
    """
    A product-like object to check status using remote metadata. Since it
    partially conforms to the Product API, it can use the same Metadata
    implementation (like File). This is used to determine whether a
    task should be executed or downloaded from remote storage.

    Parameters
    ----------
    file_ : ploomber.products.File
        Product to check status

    Notes
    -----
    Must be used in a context manager
    """
    def __init__(self, file_):
        self._local_file = file_
        self._metadata = Metadata(self)

        self._is_outdated_status_local = None
        self._is_outdated_status_remote = None
        self._exists = None

    def _fetch_remote_metadata(self):
        if self.exists() and not self._metadata._did_fetch:
            self._local_file.client.download(
                self._local_file._path_to_metadata,
                destination=self._path_to_metadata)

            # load from values from file
            self._metadata._fetch()

            try:
                self._path_to_metadata.unlink()
            except FileNotFoundError:
                pass

    def exists(self):
        """
        Checks if remote File exists. This is used by Metadata to determine
        whether to use the existing remote metadat (if any) or ignore it: if
        this returns False, remote metadata is ignored even if it exists
        """
        if self._exists is None:
            # TODO remove checking if file exists and just make the API
            # call directly
            self._exists = (self._local_file.client is not None
                            and self._local_file.client._remote_exists(
                                self._local_file._path_to_metadata)
                            and self._local_file.client._remote_exists(
                                self._local_file._path_to_file))

        return self._exists

    def fetch_metadata(self):
        return _fetch_metadata_from_file_product(self, check_file_exists=False)

    @property
    def metadata(self):
        self._fetch_remote_metadata()
        return self._metadata

    @property
    def _path_to_metadata(self):
        """
        Path to download remote metadata
        """
        name = f'.{self._local_file._path_to_file.name}.metadata.remote'
        return self._local_file._path_to_file.with_name(name)

    def _reset_cached_outdated_status(self):
        self._is_outdated_status = None

    def _is_equal_to_local_copy(self):
        """
        Check if local metadata is the same as the remote copy
        """
        return self._local_file.metadata == self.metadata

    # TODO: _is_outdated, _outdated_code_dependency and
    # _outdated_data_dependencies are very similar to the implementations
    # in Product, check what we can abstract to avoid repetition

    def _is_outdated(self, with_respect_to_local, outdated_by_code=True):
        """
        Determines outdated status using remote metadata, to decide
        whether to download the remote file or not

        with_respect_to_local : bool
            If True, determines status by comparing timestamps with upstream
            local metadata, otherwise it uses upstream remote metadata
        """
        if with_respect_to_local:
            if self._is_outdated_status_local is None:
                self._is_outdated_status_local = self._check_is_outdated(
                    with_respect_to_local, outdated_by_code)
            return self._is_outdated_status_local
        else:
            if self._is_outdated_status_remote is None:
                self._is_outdated_status_remote = self._check_is_outdated(
                    with_respect_to_local, outdated_by_code)
            return self._is_outdated_status_remote

    def _check_is_outdated(self, with_respect_to_local, outdated_by_code):
        oudated_data = self._outdated_data_dependencies(with_respect_to_local)
        outdated_code = (outdated_by_code and self._outdated_code_dependency())
        return oudated_data or outdated_code

    def _outdated_code_dependency(self):
        """
        Determine if the source code has changed by looking at the remote
        metadata
        """
        outdated, _ = self._local_file.task.dag.differ.is_different(
            a=self.metadata.stored_source_code,
            b=str(self._local_file.task.source),
            a_params=self.metadata.params,
            b_params=self._local_file.task.params.to_json_serializable(
                params_only=True),
            extension=self._local_file.task.source.extension)

        return outdated

    def _outdated_data_dependencies(self, with_respect_to_local):
        """
        Determine if the product is outdated by checking upstream timestamps
        """
        upstream_outdated = [
            self._is_outdated_due_to_upstream(up, with_respect_to_local)
            for up in self._local_file.task.upstream.values()
        ]

        # special case: if all upstream dependencies are waiting for download
        # or up-to-date, mark this as up-to-date
        if set(upstream_outdated) <= {TaskStatus.WaitingDownload, False}:
            return False

        return any(upstream_outdated)

    def __del__(self):
        if self._path_to_metadata.exists():
            self._path_to_metadata.unlink()

    def _is_outdated_due_to_upstream(self, upstream, with_respect_to_local):
        """
        A task becomes data outdated if an upstream product has a higher
        timestamp or if an upstream product is outdated
        """
        if (upstream.exec_status == TaskStatus.WaitingDownload
                or not with_respect_to_local):
            # TODO: delete ._remote will never be None
            if upstream.product._remote:
                upstream_timestamp = (
                    upstream.product._remote.metadata.timestamp)
            else:
                upstream_timestamp = None
        else:
            upstream_timestamp = upstream.product.metadata.timestamp

        if (self.metadata.timestamp is None or upstream_timestamp is None):
            return True
        else:
            more_recent_upstream = upstream_timestamp > self.metadata.timestamp

            if with_respect_to_local:
                outdated_upstream_prod = upstream.product._is_outdated()
            else:
                outdated_upstream_prod = upstream.product._is_remote_outdated(
                    True)

            return more_recent_upstream or outdated_upstream_prod

    def __repr__(self):
        return f'{type(self).__name__}({self._local_file!r})'


def _fetch_metadata_from_file_product(product, check_file_exists):
    if check_file_exists:
        file_exists = product._path_to_file.exists()
    else:
        file_exists = True

    # but we have no control over the stored code, it might be missing
    # so we check, we also require the file to exists: even if the
    # .metadata file exists, missing the actual data file means something
    # if wrong anf the task should run again
    if product._path_to_metadata.exists() and file_exists:
        content = product._path_to_metadata.read_text()

        try:
            parsed = json.loads(content)
        except json.JSONDecodeError as e:
            raise ValueError('Error loading JSON metadata '
                             f'for {product!r} stored at '
                             f'{str(product._path_to_metadata)!r}') from e
        else:
            # TODO: validate 'stored_source_code', 'timestamp' exist
            return parsed
    else:
        return None



import traceback

from ploomber.exceptions import TaskBuildError, RenderError, TaskRenderError

from papermill.exceptions import PapermillExecutionError


def exception(exc):
    """Formats an exception into a more concise traceback

    Parameters
    ----------
    """
    exceptions = [exc]

    while exc.__cause__:
        exceptions.append(exc.__cause__)
        exc = exc.__cause__

    exceptions.reverse()

    breakpoint = None

    for i, exc in enumerate(exceptions):
        if isinstance(exc, (TaskBuildError)):
            breakpoint = i
            break

    exc_hide = exceptions[breakpoint:]

    if breakpoint is not None:
        tr = _format_exception(exceptions[breakpoint - 1])
        tr = tr + '\n' + '\n'.join(f'{_get_exc_name(exc)}: {str(exc)}'
                                   for exc in exc_hide)
    else:
        # if not breakpoint, take the outermost exception and show it.
        # this ensures we show the full traceback in case there are chained
        # exceptions
        tr = _format_exception(exceptions[-1])

    return tr


def _get_exc_name(exc):
    return f'{exc.__module__}.{type(exc).__name__}'


def _format_exception(exc):
    if isinstance(exc,
                  (PapermillExecutionError, RenderError, TaskRenderError)):
        tr = str(exc)
    else:
        tr = ''.join(
            traceback.format_exception(type(exc),
                                       exc,
                                       exc.__traceback__,
                                       limit=None))

    return tr



from pathlib import Path

from ploomber.util.dotted_path import load_dotted_path


class EntryPoint:
    """
    Handles common operations on the 4 types of entry points. Exposes a
    pathlib.Path-like interface
    """
    Directory = 'directory'
    Pattern = 'pattern'
    File = 'file'
    DottedPath = 'dotted-path'
    ModulePath = 'module-path'

    def __init__(self, value):
        self.value = value
        self.type = find_entry_point_type(value)

    def exists(self):
        if self.type == self.Pattern:
            return True
        elif self.type in {self.Directory, self.File}:
            return Path(self.value).exists()
        elif self.type == self.DottedPath:
            return load_dotted_path(self.value, raise_=False) is not None

    def is_dir(self):
        return self.type == self.Directory

    @property
    def suffix(self):
        return None if self.type not in {self.File, self.ModulePath} else Path(
            self.value).suffix

    def __repr__(self):
        return repr(self.value)

    def __str__(self):
        return str(self.value)


# TODO: this only used to override default behavior
# when using the jupyter extension, but they have to be integrated with the CLI
# to provide consistent behavior. The problem is that logic implemented
# in _process_file_dir_or_glob and _process_factory_dotted_path
# also contains some CLI specific parts that we don't require here
def find_entry_point_type(entry_point):
    """
    Step 1: If not ENTRY_POINT is defined nor a value is passed, a default
    value is used (pipeline.yaml for CLI, recursive lookup for Jupyter client).
    If ENTRY_POINT is defined, this simply overrides the default value, but
    passing a value overrides the default value. Once the value is determined.

    Step 2: If value is a valid directory, DAG is loaded from such directory,
    if it's a file, it's loaded from that file (spec), finally, it's
    interpreted as a dotted path
    """
    type_ = try_to_find_entry_point_type(entry_point)

    if type_:
        return type_
    else:
        if Path(entry_point).suffix in {'.yaml', '.yml'}:
            raise ValueError(
                'Could not determine the entry point type from value: '
                f'{entry_point!r}. The file does not exist.')
        else:
            raise ValueError(
                'Could not determine the entry point type from value: '
                f'{entry_point!r}. Expected '
                'an existing file with extension .yaml or .yml, '
                'existing directory, glob-like pattern '
                '(i.e., *.py) or dotted path '
                '(i.e., module.sub_module.factory_function).')


def try_to_find_entry_point_type(entry_point):
    if entry_point is None:
        return None
    elif '*' in entry_point:
        return EntryPoint.Pattern
    elif '::' in entry_point:
        return EntryPoint.ModulePath
    elif Path(entry_point).exists():
        if Path(entry_point).is_dir():
            return EntryPoint.Directory
        else:
            return EntryPoint.File
    elif '.' in entry_point and Path(entry_point).suffix not in {
            '.yaml', '.yml'
    }:
        return EntryPoint.DottedPath



import json
from collections.abc import Mapping
import logging
from pathlib import Path
from reprlib import Repr

import yaml
from jinja2 import (Environment, Template, UndefinedError, FileSystemLoader,
                    PackageLoader, StrictUndefined)

from ploomber.exceptions import RenderError, SourceInitializationError
from ploomber.placeholders import util, extensions, abc


class Placeholder(abc.AbstractPlaceholder):
    """
    Placeholder powers all the objects that use placeholder variables (
    between curly brackets). It uses a jinja2.Template object under the hood
    but adds a few important things:

    * Keeps the raw (undendered) value: template.raw
    * Keeps path to raw value: template.location
    * Strict: will not render if missing or extra parameters
    * Upon calling .render, saves the rendered value for later access

    End users should not manipulate Placeholder objects, they should be
    automatically created from strings, pathlib.Path or jinja2.Template
    objects.

    Placeholder is mostly used by sources whose source code are parametrized
    strings (e.g. SQL scripts)

    Parameters
    ----------
    hot_reload : bool, optional
        Makes the placeholder always read the template from the file before
        rendering

    required
        Set of keys required for rendering

    Attributes
    ----------
    variables : set
        Returns the set of variables in the template (values sourrounded by
        {{ and }})

    path : pathlib.Path
        The location of the raw object. None if initialized with a str or with
        a jinja2.Template created from a str

    Notes
    -----
    You can use "raise" in a placeholder to raise exceptions, useful for
    validating input parameters: "{% raise 'some error message' %}"
    """

    def __init__(self, primitive, hot_reload=False, required=None):
        self._logger = logging.getLogger('{}.{}'.format(
            __name__,
            type(self).__name__))
        self._hot_reload = hot_reload

        self._variables = None
        self.__template = None

        # we have to take care of 4 possible cases and make sure we have
        # all we need to initialize the template, this includes having
        # access to the raw template (str) and a way to re-initialize
        # the jinja.environment.loader object (to make sure copies and
        # pickles work)

        if isinstance(primitive, Path):
            self._path = primitive
            self.__raw = primitive.read_text()
            self._loader_init = None
        elif isinstance(primitive, str):
            self._path = None
            self.__raw = primitive
            self._loader_init = None

        elif isinstance(primitive, Template):
            # NOTE: primitive.filename will be '<template>' if Template was
            # loaded from a string
            path = Path(primitive.filename)

            if primitive.environment.undefined != StrictUndefined:
                raise ValueError('Placeholder can only be initialized '
                                 'from jinja2.Templates whose undefined '
                                 'parameter is set to '
                                 'jinja2.StrictUndefined, set it explicitely '
                                 'either in the Template or Environment '
                                 'constructors')

            # we cannot get the raw template on this case, raise error
            # check '<template>' first, because Path('<template>').exists()
            # breaks on windows
            if primitive.filename == '<template>' or not path.exists():
                raise ValueError(
                    'Could not load raw source from '
                    'jinja2.Template. This usually happens '
                    'when the placeholder is initialised with a '
                    'jinja.Template which was initialized with '
                    'a string. Only jinja2.Templates loaded from '
                    'the filesystem are supported. Use '
                    'ploomber.SourceLoader or jinja\'s '
                    'FileSystemLoader/PackageLoader to fix '
                    'this issue, if you want to create a template from '
                    'a string pass it directly '
                    'Placeholder("some {{placeholder}}")')

            self._path = path
            self.__raw = path.read_text()
            self._loader_init = _make_loader_init(primitive.environment.loader)
        # SourceLoader returns Placeholder objects, which could inadvertedly
        # be passed to another Placeholder constructor when instantiating
        # a source object, since they sometimes use placeholders
        #  make sure this case is covered
        elif isinstance(primitive, Placeholder):
            self._path = primitive.path
            self.__raw = primitive._raw
            self._loader_init = _make_loader_init(
                primitive._template.environment.loader)
        else:
            raise TypeError('{} must be initialized with a Template, '
                            'Placeholder, pathlib.Path or str, '
                            'got {} instead'.format(
                                type(self).__name__,
                                type(primitive).__name__))

        if self._path is None and hot_reload:
            raise ValueError('hot_reload only works when Placeholder is '
                             'initialized from a file')

        # TODO: remove
        self.needs_render = self._needs_render()

        self._str = None if self.needs_render else self._raw
        self._repr = Repr()
        self._repr.maxstring = 40

        if required:
            self._validate_required(required)

    def _validate_required(self, required):
        missing_required = set(required) - self.variables

        if missing_required:
            msg = ('The following tags are required. ' +
                   display_error(missing_required, required))
            raise SourceInitializationError(msg)

    @property
    def _template(self):
        if self.__template is None or self._hot_reload:
            self.__template = _init_template(self._raw, self._loader_init)

        return self.__template

    @property
    def _raw(self):
        """A string with the raw jinja2.Template contents
        """
        if self._hot_reload:
            self.__raw = self._path.read_text()

        return self.__raw

    @_raw.setter
    def _raw(self, value):
        self.__raw = value

    @property
    def path(self):
        return self._path

    def _needs_render(self):
        """
        Returns true if the template is a literal and does not need any
        parameters to render
        """
        env = self._template.environment

        # check if the template has the variable or block start string
        # is there any better way of checking this?
        needs_variables = (env.variable_start_string in self._raw
                           and env.variable_end_string in self._raw)
        needs_blocks = (env.block_start_string in self._raw
                        and env.block_end_string in self._raw)

        return needs_variables or needs_blocks

    def __str__(self):
        if self._str is None:
            raise RuntimeError('Tried to read {} {} without '
                               'rendering first'.format(
                                   type(self).__name__, repr(self)))

        return self._str

    def render(self, params, optional=None, required=None):
        """
        """
        optional = optional or set()
        optional = set(optional)

        passed = set(params.keys())
        available = passed | set(self._template.environment.globals)

        missing = self.variables - available
        extra = passed - self.variables - optional

        # FIXME: self.variables should also be updated on hot_reload
        if missing:
            raise RenderError('in {}, missing required '
                              'parameters: {}, params passed: {}'.format(
                                  repr(self), missing, params))

        if extra:
            raise RenderError('in {}, unused parameters: {}, params '
                              'declared: {}'.format(repr(self), extra,
                                                    self.variables))

        try:
            self._str = self._template.render(**params)
        except UndefinedError as e:
            # TODO: we can use e.message to see which param caused the
            # error
            raise RenderError('in {}, jinja2 raised an UndefinedError, this '
                              'means the template is using an attribute '
                              'or item that does not exist, the original '
                              'traceback is shown above. For jinja2 '
                              'implementation details see: '
                              'http://jinja.pocoo.org/docs/latest'
                              '/templates/#variables'.format(
                                  repr(self))) from e

        return str(self)

    def best_repr(self, shorten):
        """
        Returns the rendered version (if available), otherwise the raw version
        """
        best = self._raw if self._str is None else self._str

        if shorten:
            best = self._repr.repr(best)

        return best

    @property
    def variables(self):
        """Returns declared variables in the template
        """
        # this requires parsing the raw template, do lazy load, but override
        # it if hot_reload is True
        if self._variables is None or self._hot_reload:
            self._variables = util.get_tags_in_str(self._raw)

        return self._variables

    def __repr__(self):
        content = self.best_repr(shorten=True)
        return f'{type(self).__name__}({content})'

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_logger']
        del state['_Placeholder__template']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._logger = logging.getLogger('{}.{}'.format(
            __name__,
            type(self).__name__))

        self.__template = None


def _init_template(raw, loader_init):
    """
    Initializes template, taking care of configuring the loader environment
    if needed

    This helps prevents errors when using copy or pickling (the copied or
    unpickled object wont have access to the environment.loader which breaks
    macros and anything that needs access to the jinja environment.loader
    object
    """
    if loader_init is None:
        template = Template(raw,
                            undefined=StrictUndefined,
                            extensions=(extensions.RaiseExtension, ))
        _add_globals(template.environment)
        return template
    else:
        if loader_init['class'] == 'FileSystemLoader':
            loader = FileSystemLoader(**loader_init['kwargs'])
        elif loader_init['class'] == 'PackageLoader':
            loader = PackageLoader(**loader_init['kwargs'])
        else:
            raise TypeError('Error setting state for Placeholder, '
                            'expected the loader to be FileSystemLoader '
                            'or PackageLoader')

        env = Environment(loader=loader,
                          undefined=StrictUndefined,
                          extensions=(extensions.RaiseExtension, ))
        _add_globals(env)

        return env.from_string(raw)


def _add_globals(env):
    """Update jinja2.Template.environment to add utility globals
    """
    env.globals['get_key'] = _get_key


def _get_key(path, key):
    path = Path(path)

    suffix2fn = {'.json': json.loads, '.yaml': yaml.safe_load}

    fn = suffix2fn.get(path.suffix)

    if not fn:
        raise ValueError('get_key must be used with .json or .yaml files. '
                         f'Got: {path.suffix!r}')

    return fn(path.read_text())[key]


def _get_package_name(loader):
    # the provider attribute was introduced in jinja 2.11.2, previous
    # versions have package_name
    if hasattr(loader, 'package_name'):
        return loader.package_name
    else:
        return loader.provider.loader.name


def _make_loader_init(loader):
    if loader is None:
        return None
    if isinstance(loader, FileSystemLoader):
        return {
            'class': type(loader).__name__,
            'kwargs': {
                'searchpath': loader.searchpath
            }
        }
    elif isinstance(loader, PackageLoader):
        return {
            'class': type(loader).__name__,
            'kwargs': {
                'package_name': _get_package_name(loader),
                'package_path': loader.package_path
            }
        }
    else:
        raise TypeError('Only templates with loader type '
                        'FileSystemLoader or PackageLoader are '
                        'supported, got: {}'.format(type(loader).__name__))


class SQLRelationPlaceholder(abc.AbstractPlaceholder):
    """
    Sources are also used by some Products that support placeholders, such as
    SQLRelation or File.

    An structured Placeholder to represents a database relation (table or
    view). Used by Products that take SQL relations as parameters.
    Not meant to be used directly by end users.

    Parameters
    ----------
    source: tuple
      A (schema, name, kind) or a (name, kind) tuple, where kind is either
      'table' or 'view'


    Notes
    -----
    Beware that SQLRelationPlaceholder(('schema', 'data', 'table')) is
    different from SQLRelationPlaceholder(('"schema"', '"data"', '"table"')).
    The first one will be rendered as schema.data and the second one as
    "schema"."data" (in SQL, quoted identifiers are case-sensitive)
    """

    # TODO: allow templating in schema

    def __init__(self, source):
        if len(source) == 2:
            name, kind = source
            schema = None
        elif len(source) == 3:
            schema, name, kind = source
        else:
            raise ValueError(
                f'{type(self).__name__} must be initialized a collection '
                f'of 2 or 3 elements, got: {source!r} '
                f'({len(source)} elements). '
                'Example: ["schema", "name", "table"]')

        # ignore empty string
        if schema == '':
            schema = None

        if name is None:
            raise ValueError('name cannot be None')

        if kind not in ('view', 'table'):
            raise ValueError('kind must be one of ["view", "table"] '
                             'got "{}"'.format(kind))

        self._schema = schema
        self._name_template = Placeholder(name)
        self._kind = kind

        # if source is literal, rendering without params should work, this
        # allows this template to be used without having to render the dag
        # first
        if not self._name_template.needs_render:
            self._name_template.render({})

    @property
    def schema(self):
        return self._schema

    @property
    def name(self):
        return str(self._name_template)

    @property
    def kind(self):
        return self._kind

    # FIXME: THIS SHOULD ONLY BE HERE IF POSTGRES

    def _validate_name(self, name):
        if len(name) > 63:
            url = ('https://www.postgresql.org/docs/current/'
                   'sql-syntax-lexical.html#SQL-SYNTAX-IDENTIFIERS')
            raise ValueError('"{}" exceeds maximum length of 63 '
                             ' (length is {}), '
                             'see: {}'.format(name, len(name), url))

    def render(self, params, **kwargs):
        name = self._name_template.render(params, **kwargs)
        self._validate_name(name)
        return str(self)

    def _qualified_name(self, unrendered_ok, shorten):
        qualified = ''

        if self.schema is not None:
            qualified += self.schema + '.'

        if unrendered_ok:
            qualified += self._name_template.best_repr(shorten=shorten)
        else:
            qualified += str(self._name_template)

        return qualified

    def __str__(self):
        return self._qualified_name(unrendered_ok=False, shorten=False)

    def _raw_repr(self):
        name = self._name_template.best_repr(shorten=False)

        elements = (name,
                    self.kind) if self.schema is None else (self.schema, name,
                                                            self.kind)

        return repr(elements)

    def __repr__(self):
        return f'{type(self).__name__}({self._raw_repr()})'

    def best_repr(self, shorten):
        return self._qualified_name(unrendered_ok=True, shorten=shorten)

    def __hash__(self):
        return hash((self.schema, self.name, self.kind))


def display_error(keys, descriptions):
    if isinstance(descriptions, Mapping):
        msg = '\n'

        for key, error in descriptions.items():
            msg += '"{key}": {error}\n'.format(key=key, error=error)

        return msg

    else:
        return ', '.join(keys)
